;=============================================================================
;KONSOLEN-TREIBER (Tastatur+2 Bildschirme)
;=============================================================================
;Umschaltflags hchstwertiges Byte -> niederwertiges Byte, x nicht belegt
;Bit 10:print/sysrq  
;Bit 9:pause/break 
;Bit 8:alt-r 
;Bit 7:alt-l 
;Bit 6:ctrl-r 
;Bit 5:ctrl-l  
;Bit 4:shift-r 
;Bit 3:shift-l 
;Bit 2:caps-lock 
;Bit 1:num-lock (Bei XMOS: Konsole 1/2 select)
;Bit 0:scroll-lock

equ keyboard_data,$7ff6
equ keyboard_ctrl,$7ff7          
equ xgraph,$7fe8

;Konsolentreiber initialisieren
;-----------------------------------------------------------------------------
konsole:	move	   #$20,@xgraph+6	;CON 1 Sichtbar
		clr	   vt52_seq		;Nicht im VT52-Mode
		clr	   vt52_seq2
		clr	   con_in_use
		clr	   active_con
		clr        umschaltflags	;Sondertasten
	
		push	   #.driver0		;Treiber anmelden
		jsr	   NEWDRIVER
		push	   #.driver1
		jsr	   NEWDRIVER
		push	   #.error		;Treiber fr Fehlermeldungen
		jsr	   NEWDRIVER
		add	   #3,sp,=

		push	   #.initstr1		;Initstrings ausgeben
		push	   #3
		jsr	   .driver1+5
		push	   #.initstr0
		push	   #3
		jsr	   .driver0+5
		add	   #4,sp

;Tastaturpuffer init
		push	   #32			;32-Worte-Ringpuffer
		jsr	   MALLOC
		inc	   sp
		sub	   -,d0,d1
		sub	   -,-,(d1)		;Systemspeicherblock
		move       d0,t_buffer		;Empfangspuffer
		add        #16,d0,t_buffer_end	;soll 16 Zeichen enthalten
		move	   t_buffer_end,t_buffer2
		add        #32,d0,t_buffer_end2

.buffer_clear:  sub	   -,-,(d0)		;#$ffff Schreiben
		inc        d0			;nchstes Wort
		cmp        d0,t_buffer_end2	;Bis alle Worte in beiden Puffern...
		jne        .buffer_clear	;... voll sind

		move       t_buffer,t_read_ptr	;Lesezeiger init
		move       t_buffer,t_write_ptr	;Schreibzeiger init
		move       t_buffer2,t_read_ptr2
		move       t_buffer2,t_write_ptr2	

.keyonoff:	in	   keyboard_data,-	;Empfangsregister lschen
		move	   #.tastatur_aus,1	;IRQ1-Vektor einstellen
		move	   #.t_commands,t_int_ar2 ;Zeiger auf auszugebende Bytes

		out	   #0,keyboard_ctrl	;Tastatur zurcksetzen
		add	   #10,timerl,d0	;(Aus-und-Anschalten)
.delay1:	cmp	   timerl,d0
		jpl	   .delay1
		in	   keyboard_data,-	;Empfangsregister lschen
		out	   #2,keyboard_ctrl	;Spannung an

		add	   #16,timerl,d0
.wait_init:	cmp	   timerl,d0		;Timeout?
		jmi	   .keyonoff		;Ja, Tastatur nochmal reseten
		cmp	   #.done,1		;Fertig?
		jne	   .wait_init		;Noch nicht
		rts

.initstr0:	dw	   27,"E",27,"j",27,"b2",27,"c0Konsole 0 bereit.",13,10,0
.initstr1:	dw	   27,"E",27,"j",27,"b1",27,"c7Konsole 1 bereit.",13,10,0

;IRQ-Routine: 11-Bit-Modus beibehalten
.retry:		move    sf flags1,~dummy	;Interruptflag lschen
.tastatur_aus:  out	   #3,keyboard_ctrl	;Clockleitung auf low
		or         ~-,#512,flags1	;Flags sichern  
		add	   #2,timerl,t_int_ar	;kurz warten
.delay2:	cmp	   timerl,t_int_ar
		jpl	   .delay2
		and        @keyboard_data,#$ff,t_int_ar ;Tastaturbyte holen
		out        #2,keyboard_ctrl	  ;Clockleitung auf tri-state
		cmp        #$aa,t_int_ar	  ;'alles ok' Byte?
		jne        .retry		  ;Nein, nochmal versuchen

;IRQ-Routine: Kommandobytes senden
.next_byte:     cmp        (t_int_ar2),-		;Fertig?
		jeq        .lamps			;ja, zur normalen IRQ-Routine 
		out        (t_int_ar2),keyboard_data	;Kommando senden
		inc        t_int_ar2			;Zeiger inc
		move    sf flags1,~dummy		;Interruptflag lschen
		or         ~-,#512,flags1		;Flags sichern  
		and        @keyboard_data,#$ff,t_int_ar ;Byte holen
		cmp        #$fa,t_int_ar		;Letztes Byte ok?
		jeq        .next_byte			;Ja, nchstes
		cmp        t_int_ar,#$fe		;Letztes Byte nochmal?
		jne        .retry			;nein, ALLES nochmal

		and     sf (t_int_ar2),#$8000,t_int_ar	;MSB betrachten
		dec  ne    t_int_ar2			;Zeiger decrementieren
		dec        t_int_ar2			;Zeiger decrementieren
		jmp        .next_byte   

;Initialisierungskommandos
.t_commands:	dw      $f0,32768+3	;Code-Set 3
		dw      $fc,32768+57	;ALT-GR auf Make/Break
		dw      $fc,32768+88	;Rechte CTRL auf Make/Break
		dw      $fb,32768+111	;Page-up auf typematic
		dw      $fb,32768+109	;Page-down auf typematic
		dw      $fd,32768+124	;Groe +Taste auf Make
		dw      $fd,32768+20	;CapsLock auf Make
		dw	$f3,32768	;Schnelles Typematic
		dw      $f4		;Arbeit wieder aufnehmen
		dw      0		;Endekennung

;Normale Tastatur-Interruptroutine IRQ-Level 1
;-----------------------------------------------------------------------------
.lamps:		out        #$ed,keyboard_data		;Lampen-Kommando senden
		move    sf flags1,~dummy 		;Umschalten in User Mode
		or         ~-,#512,flags1		;Interruptflag lschen
		and        @keyboard_data,#$ff,t_int_ar ;Byte holen
		cmp        t_int_ar,#$fe		;Nicht verstanden?
		jeq        .lamps			;Kommando wiederholen
		and        #7,umschaltflags,@keyboard_data ;Lampen einstellen
		move    sf flags1,~dummy		;Umschalten in User Mode
		or         ~-,#512,flags1		;Interruptflag lschen
		and        @keyboard_data,#$ff,t_int_ar ;Byte holen
		cmp        t_int_ar,#$fe		;Nicht verstanden?
		jeq        .lamps			;Kommando wiederholen
.t_clr_int:     move    sf flags1,~dummy		;Umschalten in User Moden
.done:		or         ~-,#512,flags1               ;Flags sichern  
		and        @keyboard_data,#$ff,t_int_ar ;Zeichen holen
		cmp        t_int_ar,#$f0		;Breakcode?
		jeq        .t_clr_int			;ja, ignorieren
		cmp        t_int_ar,#20		;caps-lock?     
		jeq        .caps_lock
		cmp        t_int_ar,#18		;shift-links?   
		jeq        .shift_links
		cmp        t_int_ar,#89		;shift-rechts?  
		jeq        .shift_rechts
		cmp        t_int_ar,#25		;alt-links?     
		jeq        .alt_links
		cmp        t_int_ar,#57		;alt-rechts?    
		jeq        .alt_rechts
		cmp        t_int_ar,#17		;ctrl-links?    
		jeq        .ctrl_links
		cmp        t_int_ar,#88		;ctrl-rechts?   
		jeq        .ctrl_rechts
		cmp        t_int_ar,#95		;scroll-lock?   
		jeq        .scroll_lock
		cmp        t_int_ar,#98		;pause/break?   
		jeq        .pause_break
		cmp        t_int_ar,#87		;print/sysrq?   
		jeq        .print_sysrq
		cmp        t_int_ar,#118	;num-lock?      
		jeq        .num_lock

;Es ist keine Sondertaste, sondern eine Zeichenerzeugende Taste
		bswp       -,t_int_ar,t_int_ar2		;Tastencode in AR2 merken
		and     sf #508,umschaltflags,-
		add  eq    #.tabelle-7,t_int_ar		;keine Sondertaste
		jeq        .zeichen_holen
		and     sf #28,umschaltflags,-          ;shift od. capslock?
		add  ne    #.tab_shift-7,t_int_ar	;ja
		jne        .zeichen_holen
		and     sf #96,umschaltflags,-          ;ctrl?
		add  ne    #.tab_ctrl-7,t_int_ar	;ja
		jne        .zeichen_holen
		add        #.tab_alt-7,t_int_ar		;sonst:Alt

.zeichen_holen: or         (t_int_ar),t_int_ar2		;Tastaturpuffereintrag berechnen
		lsr	   umschaltflags,-,t_int_ar	;Konsole 1 aktiv?
		and	   -,t_int_ar
		cmp	   active_con,t_int_ar
		jne	   .satz2			;Ja, im 2. Tastaturpuffer eintragen
		add	sf -,(t_write_ptr),-		;Puffer frei?
		jne        .t_clr_int			;nein, fertig
		move	   t_int_ar2,(t_write_ptr)	;Taste in Puffer eintragen
		inc        t_write_ptr                  ;Schreibzeiger inc
		cmp        t_write_ptr,t_buffer_end     ;Wrap_around?
		move eq    t_buffer,t_write_ptr         ;ja
		jmp        .t_clr_int                   ;wieder von vorn

.satz2:		add	sf -,(t_write_ptr2),-		;Puffer frei?
		jne        .t_clr_int			;nein, fertig
		move	   t_int_ar2,(t_write_ptr2)	;Taste in Puffer eintragen
		inc        t_write_ptr2			;Schreibzeiger inc
		cmp        t_write_ptr2,t_buffer_end2	;Wrap_around?
		move eq    t_buffer2,t_write_ptr2	;ja
		jmp        .t_clr_int			;wieder von vorn

.num_lock:	eor	   #2,umschaltflags		;Numlock togglen
		lsr	   umschaltflags,-,t_int_ar	;Konsole umschalten
		and	   -,t_int_ar
		or	   #$20,t_int_ar,@xgraph+6
		cmp	   t_int_ar,active_con
		jne	   .x2
		move 	   bcolor,@xgraph+6
		or 	   #$10,fcolor,@xgraph+6
		jmp	   .lamps
.x2:		move 	   bcolor2,@xgraph+6
		or 	   #$10,fcolor2,@xgraph+6
		jmp	   .lamps

.caps_lock:     eor        #4,umschaltflags		;Caps-Flag toggle
		jmp        .lamps			;und Lampen anpassen

.shift_links:   eor        #8,umschaltflags		;Flag togglen
		and        #$fffb,umschaltflags		;CapsLock aus
		jmp        .lamps			;weiter

.shift_rechts:  eor        #16,umschaltflags		;Flag togglen
		and        #$fffb,umschaltflags		;CapsLock aus
		jmp        .lamps			;weiter

.ctrl_links:    eor        #32,umschaltflags		;Flag toggeln
		jmp        .t_clr_int                   ;PC neu einstellen

.ctrl_rechts:   eor        #64,umschaltflags		;Flag toggeln
		jmp        .t_clr_int                   ;PC neu einstellen

.alt_links:     eor        #128,umschaltflags		;Flag toggeln
		jmp        .t_clr_int                   ;PC neu einstellen

.alt_rechts:    eor        #256,umschaltflags		;Flag toggeln
		jmp        .t_clr_int                   ;PC neu einstellen

.scroll_lock:   eor        -,umschaltflags		;Flag toggle
		jmp        .lamps			;PC neu einstellen

.pause_break:   eor        #512,umschaltflags		;Flag togglen
		jmp        .t_clr_int                   ;PC neu einstellen

.print_sysrq:   eor        #1024,umschaltflags		;Flag togglen
		jmp        .t_clr_int                   ;PC neu einstellen

.tabelle:   dw 0,27,0,0,0,0,9,94,0,0					;7-16
	    dw 0,0,60,0,$71,49,0,0,0,121,115,97,119,50,0,0		;17-32
	    dw 99,120,100,101,52,51,0,0,32,118,102,116,114,53,0,0	;33-48
	    dw 110,98,104,103,122,54,0,0,109,$6d,$6a,$75,$37,$38,0	;49-63
	    dw 0,$2c,$6b,$69,$6f,$30,$39,0,46,$2e,$2d,$6c,$94,$70,$e1,0 ;64-79
	    dw 0,132,$84,$23,$81,$27,0,0,0,0,$D,$2b,0,0,0,0		;80-95
	    dw 0,0,0,0,0,0,8,0,49,$31,0,$34,$37,0,0,0 			;96-111
	    dw $30,$2c,$32,$35,$36,$38,47,$2F,0,13,$33,43,$2b,$39,$2a	;112-126
	    dw 0,0,0,0,0,$2D						;127-132
.tab_shift: dw 0,27,0,0,0,0,9,248,0,0                                             ;7-16
	    dw 0,0,62,0,81,33,0,0,0,89,83,65,87,34,0,0			;17-32
	    dw 67,88,68,69,36,21,0,0,32,86,70,84,82,$25,0,0		;33-48
	    dw 78,66,72,71,90,38,0,0,77,$4d,$4a,$55,$2f,$28,0		;49-63
	    dw 0,$3b,$4b,$49,$4f,$3d,$29,0,58,$3a,$5f,$4c,$99,$50,$3f,0 ;64-79
	    dw 0,142,$8e,$60,$9a,$27,0,0,0,0,13,$2a,0,0,0,0		;80-95
	    dw 0,0,0,0,0,0,8,0,0,$31,0,$34,$37,0,0,0			;96-111
	    dw $30,$2c,$32,$35,$36,$38,47,$2f,0,13,$33,43,$2b,$39,$2a	;112-126
	    dw 0,0,0,0,0,$2d						;127-132
.tab_alt:   dw 0,27,0,0,0,0,9,94,0,0					;Scancodes 7-16
	    dw 0,0,$7c,0,$40,49,0,0,0,121,115,97,119,$fd,0,0		;17-32
	    dw 99,120,100,101,52,51,0,0,32,118,102,116,114,53,0,0	;33-48
	    dw 110,98,104,103,122,54,0,0,109,$e6,$6a,$75,$7b,$5b,0	;49-63
	    dw 0,$2c,$6b,$69,$6f,$7d,$5d,0,46,$2e,$2d,$6c,$94,$70,$5c,0 ;64-79
	    dw 0,132,$84,$23,$81,$27,0,0,0,0,$D,$7e,0,0,0,0		;80-95
	    dw 0,0,0,0,0,0,8,0,49,$31,0,$34,$37,0,0,0			;96-111
	    dw $30,$2c,$32,$35,$36,$38,47,$2F,0,13,$33,43,$2b,$39,$2a	;112-126
	    dw 0,0,0,0,0,$2D						;127-132
.tab_ctrl:  dw 0,27,0,0,0,0,9,30,0,0					;Scancodes 7-16
	    dw 0,0,60,0,17,49,0,0,0,25,19,1,23,50,0,0			;17-32
	    dw 3,24,4,5,52,51,0,0,32,22,6,20,18,53,0,0			;33-48
	    dw 14,2,8,7,26,54,0,0,109,13,10,21,$37,27,0			;49-63
	    dw 0,$2c,11,$9,15,$30,29,0,46,$2e,31,12,$94,16,28,0		;64-79
	    dw 0,132,$84,$23,$81,$27,0,0,0,0,$D,$2b,0,0,0,0		;80-95
	    dw 0,0,0,0,0,0,8,0,49,$31,0,$34,$37,0,0,0   		;96-111
	    dw $30,$2c,$32,$35,$36,$38,47,$2F,0,13,$33,43,$2b,$39,$2a	;112-126
	    dw 0,0,0,0,0,$2D						;127-132


;Konsolentreiber
;-----------------------------------------------------------------------------
.error:		dw	   "err",0		;Treibername "err"
		dw	   1			;Zeichentreiber
		and	sf #2,umschaltflags,-	;Welche Konsole ist sichtbar?
		jeq	   .semwait0		;Konsole 0
		jne	   .semwait1		;Konsole 1

.driver0: 	dw	   "con0"		;Treibername "con0"
		dw	   1			;ist ein Zeichentreiber
.semwait0:	irqoff
		move	sf con_in_use,-		;Konsole schon aktiv?
		move eq	   akt_task,con_in_use	;Falls nicht, belegen
		irqon
		jeq	   .semok0
;(Hier knnte ein Fairness-Flag gestzt werden!)
		jsr	   SWITCHTASK		;Sonst warten
		jmp	   .semwait0
.semok0:	move	   #$30,@xgraph+6	;CON 0 Schreibbar
		move	sf active_con,-		;Konsole 0 schon aktiv?
		jeq	   .driversame		;Ja
		clr	   active_con		;Konsole 0 aktivieren
		jmp	   .othercon		;Variablenfeld wechseln

.driver1: 	dw	   "con1"		;Treibername "con1"
		dw	   1			;ist ein Zeichentreiber
.semwait1:	irqoff
		move	sf con_in_use,-		;Konsole schon aktiv?
		move eq	   akt_task,con_in_use	;Falls nicht, belegen
		irqon
		jeq	   .semok1
;(Hier knnte ein Fairness-Flag gestzt werden!)
		jsr	   SWITCHTASK		;Sonst warten
		jmp	   .semwait1
.semok1:	move	   #$31,@xgraph+6	;CON 1 Schreibbar
		move	sf active_con,-		;Konsole 1 schon aktiv?
		jne	   .driversame		;Ja
		move	   -,active_con		;Konsole 1 aktivieren

.othercon:	irqoff
		move	   t_read_ptr,d0	;Variablenfeld wechseln
		move	   t_read_ptr2,t_read_ptr
		move	   d0,t_read_ptr2
		move	   t_write_ptr,d0
		move	   t_write_ptr2,t_write_ptr
		move	   d0,t_write_ptr2
		move	   t_buffer,d0
		move	   t_buffer2,t_buffer
		move	   d0,t_buffer2
		move	   t_buffer_end,d0
		move	   t_buffer_end2,t_buffer_end
		move	   d0,t_buffer_end2
		move	   fcolor,d0
		move	   fcolor2,fcolor
		move	   d0,fcolor2
		move	   bcolor,d0
		move	   bcolor2,bcolor
		move	   d0,bcolor2
		irqon
		move	   cursor_adress,d0
		move	   cursor_adress2,cursor_adress
		move	   d0,cursor_adress2
		move	   cursor_x,d0
		move	   cursor_x2,cursor_x
		move	   d0,cursor_x2
		move	   cursor_y,d0
		move	   cursor_y2,cursor_y
		move	   d0,cursor_y2
		move	   vt52_seq,d0
		move	   vt52_seq2,vt52_seq
		move	   d0,vt52_seq2
		move	   save_x,d0
		move	   save_x2,save_x
		move	   d0,save_x2
		move	   save_y,d0
		move	   save_y2,save_y
		move	   d0,save_y2
		move	   save_adr,d0
		move	   save_adr2,save_adr
		move	   d0,save_adr2

.driversame:	add	   -,sp,d1
		move	   (d1),d0		;Jobnummer holen
		inc	   d1
		move	   (d1),d1		;Parameter holen

;Job 1 "Zeichen lesen"
;..............................................................................
		dec	sf d0			;Job 1?
		jne	   .nojob1		;Nein
		move       (t_read_ptr),d0	;Zeichen lesen
		add	sf -,d0,-		;doch kein Zeichen?
		jne	   .ok
		move	   #118,d0
		ror	sf -,-,-			;N-Flag setzen
		jmp	   .drvexit
.ok:		sub        -,-,(t_read_ptr)		;Zeichen im Puffer lschen
		inc        t_read_ptr			;Lesezeiger inc
		cmp        t_read_ptr,t_buffer_end      ;Ende des Puffers?
		move eq    t_buffer,t_read_ptr		;Ja, wieder zum Anfang
		and	   #$ff,d0,d1			;Low Byte betrachten
		cmp	   d1,#26			;Ctrl-Z =EOF
		jne	   .ok2
		move eq    #105,d0
		ror	sf -,-,-
		jmp	   .drvexit
		
;Job 2 "Zeichen ausgeben"
;..............................................................................
.nojob1:	dec	sf d0		;Job 2?
		jne	   .nojob2	;Nein
		jsr	   .charout
		jmp	   .ok2

;Job 3 "String ausgeben"
;..............................................................................
.nojob2:	dec	sf d0		;Job 3?			
		jne	   .nojob3	;Nein

		push	   d2
		move	   d1,d2	;Stringpointer
.loop:		move	sf (d2),d1	;Zeichen holen
		jeq	   .strend	;Stringende?
.not_yet:	jsr	   .charout
		inc	   d2		;Pointer++
		jmp	   .loop
.strend:	cmp	   #6,vt52_seq	;Eine Grafiksequenz ist nicht unterbrechbar
		jpl	   .not_yet
		pop	   d2
		jmp	   .ok2

;Job 4 "Hole Zeichen und lasse es im Puffer"
;..............................................................................
.nojob3:	dec	sf d0			;Job 4?
		jne	   .drvexit		;Nein
		move       (t_read_ptr),d0	;Zeichen lesen
		add	sf -,d0,-		;doch kein Zeichen?
		jne	   .ok2
		ror	sf -,-,-		;Kein Zeichen: N-Flag setzen
		jmp	   .drvexit
.ok2:		clr	sf ~-			;N-Flag lschen
.drvexit:	clr	   con_in_use
;(Hier mte das Fairness-Flag berprft werden!)
		rts


;Gibt das Zeichen in d1 auf dem Bildschirm aus
;..............................................................................
.charout:	and	   #255,d1
		or	sf vt52_seq,-,-		;In einer VT52-Sequenz?
		jne	   .vt52_emu		;Ja
		sub	sf #32,d1		;Druckbares Zeichen?            
		jmi	   .special		;Nein,Spezialbehandlung
		asl        d1
		asl        d1
		asl        d1
		add        #charset,d1		;Zeichenadresse
                move       cursor_adress,d0

		move	   d0,@xgraph		;Bildschirmadresse low bergeben
		bswp	   d0,-,@xgraph+1	;und high
		bswp	   (d1),-,@xgraph+7	;8 Pixel des Zeichens ausgeben
		add	   #80,d0		;Bildschirmadresse auf nchste Zeile
		move	   d0,@xgraph		;Bildschimradresse bergeben
		bswp	   d0,-,@xgraph+1	;Bildschimradresse bergeben
		move	   (d1),@xgraph+7	;8 Pixel des Zeichens ausgeben
		inc	   d1			;Pointer im Zeichensatz incrementieren
		add	   #80,d0		;Bildschirmadresse auf nchste Zeile
		move	   d0,@xgraph		;das ganze noch 7 mal wiederholen
		bswp	   d0,-,@xgraph+1
		bswp	   (d1),-,@xgraph+7
		add	   #80,d0
		move	   d0,@xgraph
		bswp	   d0,-,@xgraph+1
		move	   (d1),@xgraph+7
		inc	   d1
		add	   #80,d0
		move	   d0,@xgraph
		bswp	   d0,-,@xgraph+1
		bswp	   (d1),-,@xgraph+7
		add	   #80,d0
		move	   d0,@xgraph
		bswp	   d0,-,@xgraph+1
		move	   (d1),@xgraph+7
		inc	   d1
		add	   #80,d0
		move	   d0,@xgraph
		bswp	   d0,-,@xgraph+1
		bswp	   (d1),-,@xgraph+7
		add	   #80,d0
		move	   d0,@xgraph
		bswp	   d0,-,@xgraph+1
		move	   (d1),@xgraph+7
		inc	   d1
		add	   #80,d0
		move	   d0,@xgraph
		bswp	   d0,-,@xgraph+1
		bswp	   (d1),-,@xgraph+7
		add	   #80,d0
		move	   d0,@xgraph
		bswp	   d0,-,@xgraph+1
		move	   (d1),@xgraph+7
		inc	   d1
		add	   #80,d0
		move	   d0,@xgraph
		bswp	   d0,-,@xgraph+1
		bswp	   (d1),-,@xgraph+7
		add	   #80,d0
		move	   d0,@xgraph
		bswp	   d0,-,@xgraph+1
		move	   (d1),@xgraph+7
		inc	   d1
		add	   #80,d0
		move	   d0,@xgraph
		bswp	   d0,-,@xgraph+1
		bswp	   (d1),-,@xgraph+7
		add	   #80,d0
		move	   d0,@xgraph
		bswp	   d0,-,@xgraph+1
		move	   (d1),@xgraph+7
		inc	   d1
		add	   #80,d0
		move	   d0,@xgraph
		bswp	   d0,-,@xgraph+1
		bswp	   (d1),-,@xgraph+7
		add	   #80,d0
		move	   d0,@xgraph
		bswp	   d0,-,@xgraph+1
		move	   (d1),@xgraph+7

.one_right:	inc        cursor_adress
		dec     sf cursor_x             ;CursorX luft rckwrts (79=linker Rand, 0=rechter Rand)
                jpl	   .draw_cursor         ;Kein Zeilenumbruch
		move       #79,cursor_x         ;Cursor an Anfang neue Zeile 
		sub	   #80,cursor_adress
.one_down:	add        #1280,cursor_adress
		dec     sf cursor_y             ;(24=oberer Rand, 0=unterer Rand)
                jpl	   .draw_cursor         ;Kein Scrolling

.scrolllock:	and	sf -,umschaltflags,-	;Scroll-lock?
		jeq	   .scrollok
		jsr	   SWITCHTASK		;Ja, Taskwechsel!
		jmp	   .scrolllock

.scrollok:	clr	   @xgraph		;Bildschirm um 16 Zeilen nach oben schieben
		clr	   @xgraph+1
		clr	   @xgraph+2
		move	   #5,@xgraph+3
		move	   #255,@xgraph+4	;32000-1280 Bytes kopieren
		move	   #119,@xgraph+5
.scroll:	and	sf @xgraph,#1,-		;Warten...
		jne	   .scroll
		move	   #320,d0		;Letzte Zeile lschen
.clrline:	clr	   @xgraph+7
		clr	   @xgraph+7
		clr	   @xgraph+7
		clr	   @xgraph+7
		dec	sf d0
		jne	   .clrline
		clr	   cursor_y		;Cursor auf letzte Zeile
		sub	   #1280,cursor_adress

;Invertiert die Cursorlinie an aktueller Cursorposition
;..............................................................................
.draw_cursor:	add	   #1200,cursor_adress,d0
		move	   d0,@xgraph
		bswp	   d0,-,@xgraph+1
		move	   d0,@xgraph+2
		bswp	   d0,-,@xgraph+3
		nor	   @xgraph+7,-,@xgraph+7
		rts

;Cursorbewegungen
;..............................................................................
.cursor_right:	jsr	   .draw_cursor		;Alten Cursor lschen
		jmp	   .one_right

.cursor_left:	jsr	   .draw_cursor		;Alten Cursor lschen
		dec	   cursor_adress
		inc	   cursor_x
		cmp	   #80,cursor_x	
		jne	   .draw_cursor		;Noch nicht am linken Rand
		clr	   cursor_x
		add	   #80,cursor_adress
.one_up:	sub	   #1280,cursor_adress	;Eine Zeile hoch
		inc	   cursor_y
		cmp	   #25,cursor_y
		jne	   .draw_cursor		;Noch nicht oben gewesen
		add	   #1280,cursor_adress	;Doch, Cursorposition nicht ndern
		dec	   cursor_y
		jmp	   .draw_cursor

.cursor_up:	jsr	   .draw_cursor
		jmp	   .one_up

.cursor_down:	jsr	   .draw_cursor
		jmp	   .one_down

;Clear-Home
;..............................................................................
.clear_screen:	clr	   @xgraph		;Bildschirm lschen (lassen)
		clr	   @xgraph+1
		clr	   @xgraph+7
		clr	   @xgraph+2
		clr	   @xgraph+3
		move	   #255,@xgraph+4
		move	   #124,@xgraph+5
.clearwait:	and	sf @xgraph,#1,-		;Warten...
		jne	   .clearwait
.cursor_home:	clr	   cursor_adress
		move	   #79,cursor_x
		move	   #24,cursor_y
		jmp	   .draw_cursor

;Zeile ab Cursorposition lschen
;..............................................................................
.clline_from_c:	push	   d2
		move	   cursor_adress,d0
		move	   #16,d2
.clfc_loopy:	move	   cursor_x,d1
		move	   d0,@xgraph
		bswp	   d0,-,@xgraph+1
.clfc_loopx:	clr	   @xgraph+7
		dec	sf d1
		jpl	   .clfc_loopx
		add	   #80,d0
		dec	sf d2
		jne	   .clfc_loopy
		pop	   d2
		jmp	   .draw_cursor

;Bildschirm ab Cursorposition lschen
;..............................................................................
.clear_from_c:	push	   d2
		move	   cursor_adress,d0
		move	   #16,d2		;Aktuelle Zeile bis Ende lschen
.cfc_loopy:	move	   cursor_x,d1
		move	   d0,@xgraph
		bswp	   d0,-,@xgraph+1
.cfc_loopx:	clr	   @xgraph+7
		dec	sf d1
		jpl	   .cfc_loopx
		add	   #80,d0
		dec	sf d2
		jne	   .cfc_loopy
		pop	   d2
		sub	   #80,d0
		move	   d0,@xgraph+2
		bswp	   d0,-,@xgraph+3
		sub	   d0,#32000,d0
		move	   d0,@xgraph+4
		bswp	   d0,-,@xgraph+5
.cfc_wait:	and	sf @xgraph,#1,-		;Warten...
		jne	   .cfc_wait
		jmp	   .draw_cursor

;Bildschirm ab Cursorzeile nach unten scrollen
;..............................................................................
.scroll_down:	move	   #29440,d0		;Adresse der vorletzten Zeile
		move	   #30720,d1		;Adresse der letzten Zeile
.scrdwnloop:	cmp	   d1,cursor_adress	;Fertig?
		jpl	   .scrdwnend
		move	   d1,@xgraph
		bswp	   d1,-,@xgraph+1
		move	   d0,@xgraph+2
		bswp	   d0,-,@xgraph+3
		move	   #255,@xgraph+4
		move	   #4,@xgraph+5
.scrdwnwait:	and	sf @xgraph,#1,-		;Warten...
		jne	   .scrdwnwait
		sub	   #1280,d0
		sub	   #1280,d1		;Nchste Zeile
		jmp	   .scrdwnloop
.scrdwnend:	rts

;Spezialbehandlung fr Zeichen<=31
;..............................................................................
.special:	add	   #32,d1
		cmp	   #10,d1		;Line-Feed?
		jeq	   .cursor_down

		cmp	   #13,d1		;Carrige-Return?
		jne	   .no_cr		;Nein
		jsr	   .draw_cursor		;Ja, Alten Cursor aus
		sub	   cursor_x,#79,d0	;Cursor an Anfang der Zeile
		move	   #79,cursor_x
		sub	   d0,cursor_adress 
		jmp	   .draw_cursor		;Cursor an

.no_cr:		cmp	   #9,d1		;Tabulator?
		jne	   .no_hortab		;Nein
.next_tab:	jsr	   .cursor_right 	;Ja, sooft nach links, bis TAB erreicht
		add	   -,cursor_x,d0
		and	sf #7,d0,-
		jne	   .next_tab
		rts

.no_hortab:	cmp	   #27,d1		;ESC?
		add eq	   -,-,vt52_seq		;Nchstes Zeichen ist Steuerzeichen
		cmp	   #12,d1		;Form-Feed?
		jeq	   .clear_screen
		cmp	   #8,d1	 	;Backspace?
		jeq	   .cursor_left
		rts

;VT-52-Emulator
;..............................................................................
.vt52_emu:	cmp	   #6,vt52_seq	;Grafik?
		jmi	   .no_graph	;Nein
		move	   cursor_adress,@xgraph	;Bildschirmadresse low bergeben
		bswp	   cursor_adress,-,@xgraph+1	;und high
		move	   d1,@xgraph+7			;8 Pixel Grafikausgabe
		add	   #80,cursor_adress		;Bildschirmadresse auf nchste Zeile
		inc	   vt52_seq
		cmp	   #22,vt52_seq		;16 Bytes bertragen?
		rts ne				;Noch nicht
		clr	   vt52_seq		;Ja ESC-Sequenz fertig
		sub	   #1280,cursor_adress	;Cursor-Adresse restaurieren
		jmp	   .one_right

.no_graph:	cmp	   #2,vt52_seq		;Vordergrundfarbe ndern?
		jne	   .no_fcol
		and	   #7,d1,fcolor
		clr	   vt52_seq
		lsr	   umschaltflags,-,d0
		and	   -,d0
		cmp	   d0,active_con
		rts ne
		or   	   #$10,fcolor,@xgraph+6
		rts

.no_fcol:	cmp	   #3,vt52_seq		;Hintergrundfarbe ndern?
		jne	   .no_bcol
		and	   #7,d1,bcolor
		clr	   vt52_seq
		lsr	   umschaltflags,-,d0
		and	   -,d0
		cmp	   d0,active_con
		rts ne
		move 	   bcolor,@xgraph+6
		rts

.no_bcol:	cmp	   #4,vt52_seq		;Cursor-X-Position?
		jne	   .no_curs_x
		sub	   d1,#111,cursor_x
		inc	   vt52_seq	;Als nchstes kommt die Y-Koordinate
		rts

.no_curs_x:	cmp	   #5,vt52_seq	;Cursor-Y-Position?
		jne	   .no_curs_y
		sub	   #32,d1
		sub	   d1,#24,cursor_y
		jsr	   .draw_cursor
		clr	   cursor_adress
.pos_loop:	dec	sf d1
		jmi	   .pos_finish
		add	   #1280,cursor_adress
		jmp	   .pos_loop
.pos_finish:	sub	   cursor_x,#79,d0
		add	   d0,cursor_adress
		clr	   vt52_seq	;ESC-Sequenz beendet
		jmp	   .draw_cursor

.no_curs_y:	cmp	   #71,d1	;ESC G :Grafikzeichen
		move eq	   #6,vt52_seq	;Nchsten 16 Zeichen sind Grafikinformation
		rts eq
		cmp	   #65,d1	;ESC A: Cursor hoch
		clr eq	   vt52_seq
		jeq	   .cursor_up
		cmp	   #73,d1	;ESC I: Cursor hoch (noch nicht ganz richtig!)
		clr eq	   vt52_seq
		jeq	   .cursor_up
		cmp	   #66,d1	;ESC B: Cursor runter
		clr eq	   vt52_seq
		jeq	   .cursor_down
		cmp	   #67,d1	;ESC C: Cursor rechts
		clr eq	   vt52_seq
		jeq	   .cursor_right
		cmp	   #68,d1	;ESC D: Cursor links
		clr eq	   vt52_seq
		jeq	   .cursor_left
		cmp	   #69,d1	;ESC E: Clear Home
		clr eq	   vt52_seq
		jeq	   .clear_screen
		cmp	   #72,d1	;ESC H: Cursor Home
		jne	   .no_cur_home
		clr 	   vt52_seq
		jsr	   .draw_cursor
		jmp	   .cursor_home
.no_cur_home:	cmp	   #74,d1	;ESC J: Bildschirm ab Cursor lschen
		clr eq	   vt52_seq
		jeq	   .clear_from_c
		cmp	   #75,d1	;ESC K: Zeile ab Cursor lschen
		clr eq	   vt52_seq
		jeq	   .clline_from_c
		cmp	   #106,d1	;ESC j: Cursorposition speichern
		jne	   .no_sav_cur
		clr 	   vt52_seq
		move	   cursor_x,save_x
		move	   cursor_y,save_y
		move	   cursor_adress,save_adr
		rts
.no_sav_cur:	cmp	   #107,d1	;ESC k: Cursorposition restaurieren
		jne	   .no_res_cur
		clr 	   vt52_seq
		jsr	   .draw_cursor
		move	   save_x,cursor_x
		move	   save_y,cursor_y
		move	   save_adr,cursor_adress
		jmp	   .draw_cursor

.no_res_cur:	cmp	   #76,d1	;ESC L: Zeile einfgen
		jne	   .no_insline
		clr 	   vt52_seq
		jsr	   .draw_cursor
		jsr	   .scroll_down
		sub	   cursor_x,#79,d0	;Cursor an Anfang der Zeile
		move	   #79,cursor_x
		sub	   d0,cursor_adress 
		jmp	   .clline_from_c
.no_insline:	cmp	   #98,d1	;ESC b? :Vordergrundfarbe whlen
		move eq	   #2,vt52_seq	;Nchstes Zeichen ist Vordergrundfarbe
		rts eq
		cmp	   #99,d1	;ESC c? :Hintergrundfarbe whlen
		move eq	   #3,vt52_seq	;Nchstes Zeichen ist Vordergrundfarbe
		rts eq
		cmp	   #89,d1	;ESC Y?? :Cursorposition setzen
		move eq	   #4,vt52_seq	;Nchstes Zeichen ist Cursor-X-Position
		rts eq
		clr 	   vt52_seq
		rts

;..............................................................................
charset:	bload charset.obj
