;=============================================================================
;SCSI-Treiber
;=============================================================================
;Fehlermeldungen:
;107 = Device existiert nicht 
equ scsi_data,$7ff4 
equ scsi_ctrl,$7ff5

;-------------------------------------------------------------------------
;INITIALISIERUNG UND EINHEITEN SUCHEN
;Ermittelt alle angeschlossenen Einheit. Die Suche nach weiteren Einheiten 
;wird abgebrochen sobald sich eine Einheit nicht meldet, da heit alle 
;angeschlossenen Einheiten mssen aufeinander folgende IDs haben und mit 
;id=0 beginnen. Es wird mitgeteilt, ob eine wechselbares Medium vorliegt.
;Zustzlich wird eine Herstellerspezifische Kennung ausgegeben. 
scsi:		out        -,scsi_ctrl		;Bus Reset
		push	   #2			;kurz warten
		jsr	   DELAY
		clr        @scsi_ctrl		;Bus Reset clear
		push	   #16			;1 sec warten
		jsr	   DELAY

		move	   -,accesses		;Zugriffszhler auf 1
		clr	   removable
		clr	   not_ready_active	;Drive-not-Ready-Routine nicht aktiv
		clr	   media_writeprot	;Alle Medien beschreibbar
		clr	   scsi_in_use		;Semaphor lschen
		clr	   want_to_use_scsi	;Fairness-Flag		
		push	   #1000		;Speicher fr MDB (Media-Deskriptor-Blcke),...
		jsr	   MALLOC		;...Kommando- und Sektorpuffer
		add	   #3,sp		;Stack fr alle Aufrufe reparieren
		add	   #266,d0,media_descr	;Pointer auf MDB
		move	   d0,scsi_buffer	;Pointer auf Sektorpuffer
		add	   #256,d0,command_buffer ;Pointer auf Kommandopuffer

;.............................................................................
;Suche nach SCSI-Gerten
		clr	   d7			;Zhler lschen
		move	   #scsidriver0,d6	;Pointer auf Devicedriver0
.detect_loop:   push       d7			;Gibt es das Device?
		jsr        TEST_UNIT_READY
		inc        sp
		cmp        d0,#107		;Wenn nicht, fertig
		jeq	   .detect_end
		
		push	   d6			;SCSI-Treiber anmelden
		jsr	   NEWDRIVER
		push       scsi_buffer		;Name des Gerts holen
		push       d7      
		jsr        INQUIRY		;Herstellerinfos erfragen    
		
		push       #string_init		;"SCSI-Device" ausgeben
		push	   #3
		driver	   #konsole.error+5

		add        #$30,d7,d0		;Devicenummer ausgeben
		push       d0
		push	   #2
		driver	   #konsole.error+5

		push       #32			;Space ausgeben
		push	   #2
		driver	   #konsole.error+5
		add	   #9,sp

		move	   #14,d5		;28 Zeichen Gertenamen ausgeben
		add        #4,scsi_buffer,d4	;Pointer auf Name
.nameloop:	bswp	   (d4),-,d0
		push       d0			;Zeichen High ausgeben
		push	   #2
		driver	   #konsole.error+5	
		push       (d4)			;Zeichen Low ausgeben
		push	   #2
		driver	   #konsole.error+5
		add	   #4,sp
		inc	   d4
		dec	sf d5
		jne	   .nameloop

		and     sf (scsi_buffer),#128,-	;wechselbares Medium?
		jeq	   .nextdevice		;nein
		push       #string_ch		;"wechselbares Medium" ausgeben
		push	   #3
		driver	   #konsole.error+5
		add	   #2,sp
		add	   d7,#exponents,d0
		or	   (d0),removable		

.nextdevice:	push       #line_feed		;Zeilenvorschub ausgeben
		push	   #3
		driver	   #konsole.error+5
		add	   #2,sp

		add	   #scsidriver1-scsidriver0,d6	;nchster Treiber
		inc	   d7				;Nchstes Device
		cmp	   #8,d7
		jne        .detect_loop

.detect_end:	asl	   d7		;Ungenutzte MDB wieder freigeben
		asl	   d7		;jeder MDB belegt 8 Worte
		asl	   d7
		add	   #266,d7	;+256 Sektorpuffer +10 Kommandopuffer
		push	   d7
		push	   scsi_buffer
		jsr	   MSHRINK
		add	   #2,sp

		sub	   -,scsi_buffer,d0
		sub	   -,-,(d0)	;Speicherblock ist Systemblock

		rts

;-------------------------------------------------------------------------
;Kommandotabelle
;6-Byte-Befehle
c_test_unit_ready:      dw 0,0,0,0,0,0
c_request_sense:        dw 3,0,0,0,27,0
c_format_unit:          dw 4,0,0,0,3,0          ;Pauschal:Interleave 3
c_inquiry:              dw 18,0,0,0,36,0        ;bis zu 36 Bytes Daten werden gelesen
;10-Byte-Befehle
c_read:                 dw 40,0,0,0,0,0,0,0,1,0 ;Blockadresse mu noch angegeben werden
c_write_verify:         dw 46,0,0,0,0,0,0,0,1,0 ;Blockadresse mu noch angegeben werden 
c_read_capacity:        dw 37,0,0,0,0,0,0,0,0,0

;Ausgabestrings:
string_init:    dw "SCSI-Device ",0
string_ch:      dw " (wechselbares Medium)",0
line_feed:      dw 10,13,0

;(2^i)-Tabelle
exponents:	dw	1,2,4,8,16,32,64,128,256,512,1024,2048,4096,8192,16384,32768


;8 Devicedriver dr SCSI-Gerte
;-----------------------------------------------------------------------------
scsidriver0:	dw	"a",0,0,0		;Devicename "a"
		dw	0			;Es ist ein Blockdevice
		move	#0,d1			;Devicenumber
		jmp	driversame
scsidriver1:	dw	"b",0,0,0		;Devicename "b"
		dw	0
		move	#2-1,d1			;Devicenumber
		jmp	driversame
scsidriver2:	dw	"c",0,0,0		;Devicename "c"
		dw	0
		move	#2,d1			;Devicenumber
		jmp	driversame
scsidriver3:	dw	"d",0,0,0		;Devicename "d"
		dw	0
		move	#3,d1			;Devicenumber
		jmp	driversame
scsidriver4:	dw	"e",0,0,0		;Devicename "e"
		dw	0
		move	#4,d1			;Devicenumber
		jmp	driversame
scsidriver5:	dw	"f",0,0,0		;Devicename "f"
		dw	0
		move	#5,d1			;Devicenumber
		jmp	driversame
scsidriver6:	dw	"g",0,0,0		;Devicename "g"
		dw	0
		move	#6,d1			;Devicenumber
		jmp	driversame
scsidriver7:	dw	"h",0,0,0		;Devicename "h"
		dw	0
		move	#7,d1			;Devicenumber

driversame:	irqoff
		cmp	   scsi_in_use,- 	;SCSI belegt?
		move eq	   akt_task,scsi_in_use	;Nein, selber belegen
		irqon
		jeq	   .semok
		move	   akt_task,want_to_use_scsi	;Fairness-Flag setzen
		jsr	   SWITCHTASK
		jmp	   driversame

.semok:		add	   -,sp,d0
		pushall
		move	   d1,d7		;SCSI-ID
		move	   d0,d5
		move	   (d0),d6		;Jobnummer holen

		add	   #exponents,d1
		and	sf (d1),removable,-	;Ist Medium wechselbar?

		push	   d7
		jsr  ne	   TEST_UNIT_READY	;Ist Datentrger gewechselt worden?
		inc	   sp

;.............................................................................
;JOB 1: LESEN
		dec	sf d6			;Job "Lesen"?
		jne	   .writejob		;Nein

		add	   #3,d5,d0		;Sektor low
		push	   (d0)
		dec	   d0
		push	   (d0)			;Sektor high
		push	   d7			;Unitnumber
		jsr	   GETBLOCK		;Block lesen
		add	   #3,sp
		jmi	   .driverexit

		add	   #4,d5
		and	   #$ff,(d5),d2	;d2 ist Position des 1. Worts
		inc	   d5
		add	   (d5),d2,d3	;d3 ist Position des letzten Wortes
		cmp	   #$101,d3	;Sicherstellen, da es nicht grer als $100 ist
		move pl	   #$100,d3
		sub	   d2,d3	;d3 gibt Anzahl an
		sub	   #4,d5
		move	   (d5),d1	;User-Pufferadresse holen
		add	   #4,d0	;d0 zeigt auf Cachebufferdaten
		add	   d0,d2	;d2 zeigt auf 1. Wort im Cachepuffer
.same:		clr	   d0		;Kein Fehler
.copyloop:	dec	sf d3		;Weitere Worte kopieren?
		jmi	   .driverexit	;N
		move	   (d2),(d1)	;Daten kopieren
		inc	   d1
		inc	   d2
		jmp	   .copyloop

;.............................................................................
;JOB 2: SCHREIBEN
.writejob:	dec	sf d6			;Job "Schreiben"?
		jne	   .capjob		;Nein

		add	   #exponents,d7,d1	;Medium schreibgeschtzt?
		and	sf (d1),media_writeprot,-
		move ne	   #$0727,d0		;Ja: Fehlerrckgabe...
		jne	   .driverexit		;...und fertig

		add	   #3,d5,d0		;Sektor low
		push	   (d0)
		dec	   d0
		push	   (d0)			;Sektor high
		push	   d7			;Unitnumber
		jsr	   GETBLOCK		;Block lesen
		add	   #3,sp
		jmi	   .driverexit
		or	   #8,(d0)		;Puffer ist SchreibPuffer

		add	   #4,d5
		and	   #$ff,(d5),d1	;d1 ist Position des 1. Worts
		inc	   d5
		add	   (d5),d1,d3	;d3 ist Position des letzten Wortes
		cmp	   #$101,d3	;Sicherstellen, da es nicht grer als $100 ist
		move pl	   #$100,d3
		sub	   d1,d3	;d3 gibt Anzahl an
		sub	   #4,d5
		move	   (d5),d2	;User-Pufferadresse holen
		add	   #4,d0
		add	   d0,d1	;d1 zeigt auf 1. Wort im Cachepuffer

;Schreibzugriff auf Sektor Null? Dies wird ein High-Level-Format bedeuten.
		add	   #3,d5,d0	
		cmp	   (d0),-
		jne	   .same
		dec	   d0
		cmp	   (d0),-
		jne	   .same			

;Schreibadresse war Sektor 0 -> neuen Mediadescriptor erstellen
		push	   d1
		push	   d7
		sub	   #7,sp
		jsr	   MEDIA_CHANGED
		add	   #8,sp
		pop	   d1
		jmp	   .same

;.............................................................................
;JOB 3: KAPAZITT
.capjob:	dec	sf d6			;Job "Kapazitt ermitteln"?
		jne	   .formjob		;Nein
		 
		inc	   d5			;Zeiger auf Pufferadresse
		push	   (d5)			;Pufferadresse	
		push	   d7			;Devicenummer
		jsr	   READ_CAPACITY
		add        #2,sp
		jmp	   .driverexit

;.............................................................................
;JOB 4: FORMAT
.formjob:	dec	sf d6			;Job "Format"?
		jne	   .decbufjob		;Nein

		push	   d7			;Devicenummer
		jsr	   FORMAT_UNIT
		inc	   sp
		jmp	   .driverexit

;.............................................................................
;JOB 5: DECREMENT BUFFERS
.decbufjob:	dec	sf d6			;Job "Dec Buffers"?
		jne	   .flushjob		;Nein

		inc	   d5
		move	   (d5),d6	;Bis zu welcher Adresse freimachen?

.decloop:	and	sf #8,(efs),-	;Ist es ein Lesepuffer?
		jeq	   .delete	;Ja, lschen

		add	   #2,efs,d0
		push	   (d0)		;Block# Low
		dec	   d0
		push	   (d0)		;Block# High
		add	   #3,d0
		push	   d0		;Pufferadresse
		and	   #7,(efs),d0
		push	   d0		;SCSI-Unitnumber
		jsr	   SCSIWRITE	;Sektor schreiben
		add	   #4,sp

.delete:	add	sf #260,efs	;520 Bytes an den Heap zurckgeben
		cmp	   d6,efs
		jmi	   .decloop
		jmp	   .driverexit

;.............................................................................
;JOB 6: FLUSH
.flushjob:	dec	sf d6		;Job "Flush"?
		move ne	   #$0549,d0	;Nein, Job unbekannt...
		jne	   .driverexit	;...und fertig

		or	   #8,d7	;uns interessieren nur Schreibpuffer
		move	   efs,d3	;d3 ist Pointer auf Cache-Puffer

.flushloop:	and	   #15,(d3),d0
		cmp	   d7,d0	;Schreibpuffer fr dieses Device?
		jne	   .noflush	;N

		add	   #2,d3,d0
		push	   (d0)		;Block# Low
		dec	   d0
		push	   (d0)		;Block# High
		add	   #3,d0
		push	   d0		;Pufferadresse
		and	   #7,d7,d0
		push	   d0		;SCSI-Unitnumber
		jsr	   SCSIWRITE	;Sektor schreiben
		add	   #4,sp
		and  	   #$fff7,(d3)	;Puffer ist nur noch Lesepuffer

.noflush:	add	sf #260,d3	;Nchster Puffer
		cmp	   #bs_start,d3
		jne	   .flushloop	;Noch nicht letzter
		clr	   d0

;.............................................................................
.driverexit:	clr	   scsi_in_use
		cmp	   want_to_use_scsi,-	;Fairness-Flag berprfen
		clr	   want_to_use_scsi
		jsr ne	   SWITCHTASK
		move	sf d0,-			;Fehler aufgetreten?
		ror  ne sf -,-,-		;Wenn ja :N-Flag
		popall
		rts

;.............................................................................
;Diese Routine sorgt dafr, da ein angegebener Block in einem Cachepuffer ist
getblock:	pushall
		add	   #7,sp,d0	;Devicenumber holen
		or	   #16,(d0),d2	;Nur gltige Puffer interessieren
		inc	   d0
		move	   (d0),d3	;Sektornummer high
		inc	   d0
		move	   (d0),d4	;Sektornummer low
		ror	   -,-,d5	;Zugriffszhler des Puffers, der am lngsten nicht benutzt wurde
		move	   efs,d6	;d6 pointet auf Cachepuffer

.search:	and	sf #23,(d6),d0	;Puffer gltig fr unser Device?
		cmp	   d0,d2
		jne	   .nohit	;Nein
		add	   -,d6,d0	;d0 temporrer Pointer
		cmp	   d3,(d0)	;Sektornummern high gleich?
		jne	   .nohit	;N
		inc	   d0
		cmp	   d4,(d0)	;Sektornummern low gleich?
		jne	   .nohit	;N
		move	   d6,d7	;Adresse steht in d7
		jmp	   .cachehit

;Puffer ist uninteressant. Aber vielleicht ist es der lteste
.nohit:		add	   #3,d6,d0			
		cmp	   d5,(d0)	;Ist Puffer der lteste?
		jpl	   .notoldest	;Nein, nicht verwerfen
		move	   (d0),d5	;Bis hierhin ja...
		move  	   d6,d7	;...Adresse merken
.notoldest:	add	sf #260,d6	;Nchster Puffer
		cmp	   #bs_start,d6
		jne	   .search	;Noch nicht letzter

;Es ist ein Cachemiss aufgetreten. Nun mu der Puffer an Adresse (d7)
;fr den geforderten Sektor herhalten.
		and	   #7,d2	;Unitnumber in d2
		and	sf #8,(d7),-	;Ist es ein Schreibpuffer?
		jeq	   .getnew	;Nein, sofort Sektor laden

;Der am lngsten nicht mehr benutzte Cachepuffer war leider ein 
;Schreibpuffer -> zurckschreiben
		add	   #2,d7,d0
		push	   (d0)		;Block# Low
		dec	   d0
		push	   (d0)		;Block# High
		add	   #3,d0
		push	   d0		;Pufferadresse
		and	   #7,(d7),d0	;SCSI-Unitnumber
		push	   d0
		jsr	   SCSIWRITE	;Sektor schreiben
		add	   #4,sp

;Nun kann der angeforderte Sektor gelesen werden
.getnew:	add	   #2,d7,d0	;d0 ist temporrer Pointer
		push	   d4		;Block# low pushen
		move	   d4,(d0)	;und in Pufferverwaltung eintragen
		dec	   d0
		push	   d3		;Block# high pushen
		move	   d3,(d0)	;und in Pufferverwaltung eintragen
		add	   #3,d0
		push	   d0		;Pufferadresse pushen
		move	   d2,(d7)	;Unitnumber in Pufferverwaltung...
		push	   d2		;...und auf den Stack
		jsr	   SCSIREAD	;Block lesen
		add	   #4,sp
		jmi	   .errorend	;Bei Fehler: Ende
		or	   #16,(d7)	;Puffer ist gltig

;Der Puffer enthlt nun den gewnschten Sektor
.cachehit:	add	   #3,d7,d0		;Puffer-Zugriffszhler... 
		cmp	   accesses,(d0)	;...gleich dem Gesamtzugriffszhler?
		jeq	   .end			;Ja, Zugriffszhler nicht ndern
		inc    sf  accesses		;Gesamtzugriffszhler erhhen
		move	   accesses,(d0)	;Wird Puffer-Zugriffszhler
		jpl	   .end			;Kein Zhlerberlauf

;Falls Gesamtzugriffszhler berluft (=$8000 ist), werden 
;alle Pufferzugriffszhler durch 256 geteilt
		add	   #3,efs,d0
.divloop:	add	   #$ff,(d0)	;255 addieren, damit keine 0en entstehen
		bswp	   (d0),-,(d0)	;/256
		add	   #260,d0	;Nchster Puffer
		cmp	   #bs_start,d0
		jmi	   .divloop	;Noch nicht letzter
		move	   #$81,accesses

.end:		move	   d7,d0
		clr	sf ~-		;N-Flag lschen
.errorend:	popall
		rts

;Low-Level SCSI-Funktionen
;-----------------------------------------------------------------------------
;Diese Routine kopiert das SCSI-Kommando ab Adresse (d0) 
;in den Kommandopuffer (weil der im RAM-Bereich ist!)
;d6 sagt aus wann das Kommando fr einen Taskwechsel unterbrochen werden kann,
;da der Zugriff lange dauert.
copy_command:   move      command_buffer,d7
		move      command_buffer,d1
		move      #10,d2
.copy:          move      (d0),(d1)          
		inc       d0
		inc       d1
		dec    sf d2
		jne       .copy
		rts

test_unit_ready: pushall
		move      #c_test_unit_ready,d7	;Kommandozeiger einstellen
		sub	  -,-,d6		;Kommando wird nie unterbrochen
		jmp       command               
		
format_unit:    pushall
		move      #c_format_unit,d0	;Kommando ins RAM kopieren
		jsr       copy_command                  
		move	  #6,d6			;Kommando wird unterbrochen
		jmp       command               

inquiry:        pushall
		move      #c_inquiry,d7		;Kommandozeiger einstellen
		sub	  -,-,d6		;Kommando wird nie unterbrochen
		jmp       command       

read_capacity:  pushall
		move      #c_read_capacity,d7	;Kommandozeiger einstellen
		sub	  -,-,d6		;Kommando wird nie unterbrochen
		jmp       command               

request_sense:  pushall
		move      #c_request_sense,d7	;Kommandozeiger einstellen
		sub	  -,-,d6		;Kommando wird nie unterbrochen
		jmp       command

scsiwrite:	pushall
		move      #c_write_verify,d0	;Kommando ins RAM kopieren
		jsr       copy_command
		move	  #522,d6		;Kommando bei ByteNr. 522 unterbrechen
	        jmp	  scsiread.same		;... um Verify-Vorgang abzuwarten
		
scsiread:	pushall
		move      #c_read,d0		;Kommando ins RAM kopieren
		jsr       copy_command
		move	  #10,d6		;Kommando bei ByteNr 10 unterbrechen, (Compare)
.same:		add       #9,sp,d0		;Blocknummer in das Kommando eintragen
		add       #2,command_buffer,d1
		bswp      (d0),-,(d1)		;Bits 31-24
		inc       d1
		move      (d0),(d1)		;Bits 23-16
		inc       d0
		inc       d1
		bswp      (d0),-,(d1)		;Bits 15-8
		inc       d1
		move      (d0),(d1)		;Bits 7-0
		jmp       command               
		
;Befehlsausfhrung
;-----------------------------------------------------------------------------
;bergibt Kommando an SCSI-Device. Bei Status 2 wird die Fehlermeldung im 
;Fehlerregister error zusammengefat, in den Bits 11-8 steht der 
;Statusschlssel und in den Bits 7-0 der Statuscode. Die ankommenden 
;Bytes werden zu Worten zusammengefat, dabei ist das erste Byte, das 
;gesendet oder empfangen wird, das High-Byte.
command:        add        #7,sp,d0
		move       (d0),d1		;Unitnumber holen
		and	   #7,d1		
		move	   d1,(d0)
		inc        d0
		move       (d0),d5		;Pufferadresse holen
		clr        d3			;high/low-Byte-Zeiger init
		clr        d4			;Kontrollzhler lschen 

;Target selektieren
		add        #exponents,d1,d0	;2 hoch Unitnumber berechnen
		move	   (d0),d0
		out        d0,scsi_data		;Target selektieren

		move	   #1000,d1		;Maximal 1000
.wait_for_busy: and     sf @scsi_ctrl,#16,-	;SCSI-Bus busy?                          
		jne        .comnext		;Ja, jetzt Daten tranferieren
		dec	sf d1			;Timeoutzhler
		jne	   .wait_for_busy 

;Target nicht verfgbar, Fehler 107
		out        #0,scsi_data		;SCSI-Bus Clear
		move       #107,d0		;Fehler melden
		jmp	   .errorend		;und mit N-Flag beenden

;Haupt-Datenbertragungs-Schleife
.comnext:	cmp	   d4,d6		;Kommando unterbrechen?
		jsr eq	   SWITCHTASK		;ja
		and     sf @scsi_ctrl,#8,-	;Request?
		jeq        .comnext		;Nein, warten
		inc        d4			;Kontrollzhler inc
		and     sf @scsi_ctrl,#7,d0	;Welche Phase?
		cmp        #2,d0		;Immer noch Command-Phase?
		jne	   .nocommand		;Nein, zur nchsten Phase
		out        (d7),scsi_data  	;Kommando-Byte ausgeben
		inc        d7			;Zeiger erhhen
		jmp        .comnext

.nocommand:	cmp	   #4,d0		;Welche ist die nchste Phase?
		jeq	   .getdata		;Data-In-Phase.
		jpl	   .getstatus		;Status-In
		jmp	   .putdata		;Data-Out

.dataoutnext:	cmp	   d4,d6		;Kommando unterbrechen?
		jsr eq	   SWITCHTASK		;ja
		and     sf @scsi_ctrl,#8,-	;Request?
		jeq        .dataoutnext		;Nein, warten
		and     sf @scsi_ctrl,#7,-	;Immer noch Data-Out-Phase?
		jne	   .getstatus		;Nein, zur nchsten Phase
.putdata:  	bswp       (d5),-,@scsi_data	;high-Byte ausgeben
.dataoutreq:	and     sf @scsi_ctrl,#8,-	;Request?
		jeq        .dataoutreq		;Nein, warten
		inc	   d4			;Kontrollzhler inc
		out	   (d5),scsi_data	;low-Byte ausgeben
		inc	   d5			;Datapointer incrementieren
		jmp	   .dataoutnext

.datainnext:	and     sf @scsi_ctrl,#8,-	;Request?
		jeq        .datainnext		;Nein, warten
		and     sf @scsi_ctrl,#2,-	;Immer noch Data-In-Phase?
		jne	   .getstatus		;Nein, zur nchsten Phase
.getdata:	in	   scsi_data,d0		;Datum holen
		bswp       -,d0,(d5)		;im high-Byte ablegen
.datainreq:	and     sf @scsi_ctrl,#8,-	;Request?
		jeq        .datainreq		;Nein, warten
		and     sf @scsi_ctrl,#2,-	;Immer noch Data-In-Phase?
		jne	   .getstatus		;Nein, zur nchsten Phase
		inc        d4			;Kontrollzhler inc
      		and        @scsi_data,#$ff,d0	;Datum holen
		or         d0,(d5),=   		;im low-Byte ablegen
		inc        d5			;Datapointer incrementieren
		jmp	   .datainnext

.getstatus:	and	   @scsi_ctrl,#15,d0	;Req fr Status?
		cmp	   #14,d0
		and  	   @scsi_data,#$ff,d0 	;Datum holen
		move eq	   d0,status		;evtl als Status speichern
		and     sf @scsi_ctrl,#16,-	;Immer noch Busy?
		jne        .getstatus		;Ja, warten

;SCSI-Datentransfer ist beendet, Fehlerberprfung
		and     sf @scsi_ctrl,#32,-	;Parityerror aufgetreten?
		move ne	   #$0447,d0		;Ja, Error ausgeben
		jne        .errorend		;...und fertig

;SCSI-Status berprfen
		cmp        status,-
		jeq        .end		 	;SCSI-Status=0:alles ok
		cmp        #2,status               
		jeq        .do_req_sense 	;SCSI-Status=2:Request Sense       
		move	   #$0801,d0	 	;Sonst: Drive-Busy-Fehler

;Routinenende
.errorend:	ror	sf -,-,-		;N_Flag setzen
		jmp	   .sameexit
.end:		clr	sf ~-			;N-Flag lschen
.sameexit:	popall
		rts				;Fertig

;Status 2 ist aufgetreten, ein Request-Sense-Kommando wird abgesetzt.
.do_req_sense:  add	   #7,sp,d0
		push       scsi_buffer		;Adresse pushen
		push	   (d0)			;Unitnumber
		jsr        REQUEST_SENSE
		add        #2,sp
		add        -,scsi_buffer,d0
		and        (d0),#$ff00,d7	;Hi-Byte:Statusschlssel
		add        #5,d0
		bswp       (d0),-,d0		;Lo-Byte:Statuscode
		or         d0,d7

;Media-Changed und Drive-not-Ready behandeln
		bswp	   d7,-,d6
		cmp	   #6,d6		;UNIT-Attention?
		jne	   .no6
		jsr  	   DRIVE_NOT_READY	;Ja->Drive-not-Ready und...
		jsr  	   MEDIA_CHANGED	;..Media-Changed aufrufen
		move	   d7,d0
		jmp	   .errorend

.no6:		cmp	   #2,d6		;Drive not ready?
		jne	   .no2
		jsr  	   DRIVE_NOT_READY	;->Drive-not-Ready aufrufen
		move	   d7,d0
		jmp	   .errorend

.no2:		cmp	   #7,d6		;Writeprotect?
		add  eq	   #7,sp,d0		;Ja, Write-Protectbit setzen
		add  eq	   #exponents,(d0),d0
		or   eq	   (d0),media_writeprot	;Writeprotect-Bit setzen
		move	   d7,d0		;Fehlernummer vom Req-Sense
		jmp	   .errorend

;Media-Changed-Routine
;-----------------------------------------------------------------------------
media_changed: 	move	sf not_ready_active,-	;Laufe ich schon?
		rts  ne				;Ja, fertig
		move 	   -,not_ready_active	;Wenn nein, Semaphorbit setzen
		add	   #8,sp,d0		;Unitnumber holen
		move	   (d0),d0
		add	   #exponents,d0,d1	;Medium sei erst mal
		nor	   (d1),-,d1
		and	   d1,media_writeprot	;nicht schreibgeschtzt

;Bootsektor (Identifikation) lesen
		push       #0			;Low Adresse Bootsektor
		push       #0			;High Adresse Bootsektor
		push	   scsi_buffer		;Pufferadresse
		push	   d0			;Unitnumber
		jsr	   SCSIREAD		;Block lesen fehlerhaft?
		add  mi	   #4,sp		;Ja, fertig
		jmi	   .fertig
		jsr        SCSIWRITE		;Nein, Testschreiben um writeprotect zu ermitteln
		add	   #4,sp

;Medium-Deskriptor aufbauen
		add	   #8,sp,d0		;Unitnumber holen
		move	   (d0),d0		
		asl	   d0			;Adresse vom Medium-Deskriptor
		asl	   d0
		asl	   d0
		add	   media_descr,d0
		add	   #8,scsi_buffer,d1	;Adresse des Mediennamens
		push	   d2
		move	   #7,d2		;7 Worte (Name) kopieren
.loop:		move	   (d1),(d0)
		inc	   d0
		inc	   d1
		dec	sf d2
		jne	   .loop
		inc	   d1
		move	   (d1),(d0)		;Seriennummer kopieren
		pop	   d2
.fertig:	clr	   not_ready_active	;Semaphor lschen
		rts
			
;Drive-Not-Ready-Routine
;-----------------------------------------------------------------------------
drive_not_ready: move	sf not_ready_active,-	;Laufe ich schon?
		rts  ne				;Ja, zurck
		move 	   -,not_ready_active	;Wenn nein, Semaphorbit setzen
		push	   d2
		push	   d3

;Weitere Schreibzugriffe verhindern
		add	   #10,sp,d0		;Devicenumber holen
		move	   (d0),d0
		add	   #exponents,d0,d1
		or	   (d1),media_writeprot

;Gibt es fr dieses Device noch Schreibpuffer?
		or	   #8,d0	;Uns interessieren nur Schreibpuffer
		move	   efs,d1	;d1 ist Pointer auf Cachepuffer
		cmp	   #bs_start,d1
		jeq	   .ende	;Falls es gar keine Puffer gibt:Fertig
.loop1:		and	   #15,(d1),d2
		cmp	   d0,d2	;Ist es einer?
		jeq	   .found	;Ja, groe Panik
		add	   #260,d1	;Nchster Puffer
		cmp	   #bs_start,d1
		jne	   .loop1	;Nein, weitersuchen

		and	   #7,d0	;Uns interessiert nur die Devicenumber
		push	   d0		;d0 sichern
		add	   #$61,d0	;Devicename
		move	   d0,(command_buffer)
		add	   -,command_buffer,d0
		clr	   (d0)
		push	   command_buffer
		jsr	   OPENED_FILES	;Gibt es noch geffnete Files fr dieses Device?
		inc	   sp
		cmp	   d0,-
		pop	   d0		;d0 restaurieren
		jne	   .found	

;Es gibt keine Schreibpuffer -> Alle Puffer fr dieses Device lschen
.clearcache:	move	   efs,d1	;d1 ist Pointer auf Cachepuffer
.loop2:		and	   #7,(d1),d2
		cmp	   d0,d2	;Ist es ein interessanter Puffer?
		clr eq	   (d1)		;Ja, Puffer ungltig machen
		add	   #3,d1
		clr eq	   (d1)		;und Zugriffszhler auf 0 setzen
		add	   #257,d1	;Nchster Puffer
		cmp	   #bs_start,d1
		jne	   .loop2	;Nein, weitersuchen

;Es gibt keine Schreibpuffer, fertig
.ende:		pop	   d3
		pop	   d2
		clr	   not_ready_active	;Semaphorbit lschen
		rts

;.............................................................................
;PANIK-Routine, da Schreibpuffer oder geffnete Dateien existieren.
.warn1:		dw	13,10,"Medium ",0
.warn2:		dw	" OHNE Schreibschutz wieder in SCSI-Device ",0
.warn3:		dw	" einlegen.",13,10,"(I)gnore (O)k.",13,10,0

;Fr das nicht bereite Drive gibt es noch Schreibpuffer
.found:		and	   #7,d0
		asl	   d0			;Adresse des Media-Deskriptors
		asl	   d0
		asl	   d0
		add	   media_descr,d0,d3

.flushkey:	push	   -			;Job 1:Zeichen lesen 
		driver	   #konsole.error+5	;Tastaturpuffer leeren
		inc	   sp
		jpl	   .flushkey

		push	   #.warn1		;"Medium " ausgeben
		push	   #3
		driver	   #konsole.error+5
		add	   #2,sp

		push	   d4
		move	   d3,d4		;Adresse des Media-Deskriptors
		move	   #7,d2		;7 Worte
.nameout:	bswp	   (d4),-,d0		;Medienname ausgeben
		push	   d0
		push	   #2
		driver	   #konsole.error+5
		add	   #2,sp

		and	   #$ff,(d4),d0
		push	   d0
		push	   #2
		driver	   #konsole.error+5
		add	   #2,sp

		inc	   d4
		dec	sf d2
		jne	   .nameout
		pop	   d4

		push	   #.warn2		;" wieder in ..." ausgeben
		push	   #3
		driver	   #konsole.error+5
		add	   #2,sp
		add	   #10,sp,d0		;Devicenumber holen
		add	   #$30,(d0),d0		;in ASCII wandeln
		push	   d0
		push	   #2
		driver	   #konsole.error+5
		add	   #2,sp
		push	   #.warn3		;" einlegen..." ausgeben
		push	   #3
		driver	   #konsole.error+5
		add	   #2,sp

.waitforkey:	push	   -
		driver	   #konsole.error+5	;Zeichen von Tastatur holen
		inc	   sp
		jmi	   .waitforkey

		bswp	   d0,-,d0		;Scancode ins lowbyte
		cmp	   #67,d0		;Ignore?
		jeq	   .clearcache		;...und Cachedaten verwerfen
		cmp	   #68,d0		;Ok? 
		jne	   .waitforkey		;Nein, weiter auf Taste warten

;Der User hat O.K. gedrckt
		add	   #10,sp,d0	;Devicenumber holen
		move	   (d0),d0
		push	   #0		;Blocknumber low
		push	   #0		;Blocknumber high
		push	   scsi_buffer	;Adresse
		push	   d0		;Device
		jsr	   SCSIREAD	;Block lesen (evtl Unit-Attention)
		jsr	   SCSIREAD	;deshalb nochmal Block lesen
		jmi	   .notestwrite
		jsr  	   SCSIWRITE	;Evtl. Testschreiben, fr writeprotect
.notestwrite:	add	   #4,sp
		jmi	   .flushkey	;Bei Schreibschutz: Meldung erneut ausgebem
			
;Ist es die Diskette von vorhin?
		move	   d3,d0		;Adresse des Mediadeskriptors
		add	   #8,scsi_buffer,d1	;Adresse des Bootsektors
		move	   #7,d2
.comploop:	cmp	   (d0),(d1)		;Medienname gleich?
		jne	   .flushkey		;Nein
		inc	   d0
		inc	   d1
		dec	sf d2
		jne	   .comploop
		inc	   d1
		cmp	   (d0),(d1)		;Seriennummer gleich?
		jne	   .flushkey		;Nein

;Status quo ante wurde vom User tatschlich wieder hergestellt
		add	   #10,sp,d1		;Devicenumber holen
		add	   #exponents,(d1),d1	;Medium ist wieder 
		nor	   (d1),-,d1
		and	   d1,media_writeprot	;beschreibbar
		jmp	   .ende
