;=============================================================================
;SERIELLE SCHNITTSTELLE
;=============================================================================
;Fehler:
;1 = Parittsfehler
;2 = Stopbitfehler
;4 = Overrunfehler
;...und Kombinationen (Addition der Fehlerwerte)!

equ empfangsregister,$7ffd
equ senderegister,$7fff
equ controllregister,$7ffe
equ statusregister,$7ffc

;Schnittstellentreiber initialisieren
;-----------------------------------------------------------------------------
;Grundeinstellung: 9600 baud, 1 Stoppbit, DTR=0, RTS=0
seriell:	push	   #16			;16 Worte fr Empfangspuffer
		jsr	   MALLOC
		inc	   sp
		sub	   -,d0,d1
		sub	   -,-,(d1)		;Ist ein Systemblock, Besitzer $ffff
		move       d0,s_buffer		;Pufferanfang
		add	   #16,d0,s_buffer_end	;Pufferende
.clear_loop:	sub	   -,-,(d0)		;Empfangspuffer lschen
		inc        d0
		cmp        d0,s_buffer_end
		jne        .clear_loop
		clr        @statusregister	;Statusregister lschen
		out        #10,controllregister	;Controllregister setzen
		move       s_buffer,s_read	;Lesezeiger init
		move       s_buffer,s_write	;Schreibzeiger init
		move       #seriell_int,2	;Interruptroutine init
		push	   #.driver		;Serielltreiber anmelden
		jsr	   NEWDRIVER
		inc	   sp
		clr	   cts			;0=Sendeerlaubnis
		move	   -,rts		
		rts
  
;Seriell-Schnittstellentreiber
;-----------------------------------------------------------------------------
.driver: 	dw	   "ser",0		;Treibername "ser"
		dw	   1			;ist ein Zeichentreiber
		add	   -,sp,d1
		move	   (d1),d0		;Jobnummer nach d0 holen
		inc	   d1
		move	   (d1),d1		;Parameter nach d1 holen

;Job 1 "Zeichen aus Puffer lesen"
;.............................................................................
		dec	sf d0			;Job 1?
		jne	   .job2		;Nein

		move    sf (s_read),d0		;Zeichen lesen, setzt N-Flag falls kein Zeichen
		move mi	   #118,d0		;Kein Zeichen, fertig
		rts  mi				

		sub	   -,-,(s_read)		;Zeichen im Puffer lschen
		inc        s_read		;Lesezeiger++
		cmp        s_read,s_buffer_end	;Pufferende?
		move eq    s_buffer,s_read	;Ja, wieder auf Anfang setzen

		move    sf (s_read),-		;Zeichen lesen, setzt N-Flag falls kein Zeichen
		jpl	   .eofcheck
		cmp	   rts,-		;Wurde ein CTRL-s von uns gesendet?
		clr  	   rts
		move       #17,d1		;CTRL-q senden, da Puffer wieder leer!
		jsr  ne	   .charout		;ja, 		

.eofcheck:	and	   #$ff,d0,d1		;Niederwertiges Byte bestimmen
		cmp	   d1,#26		;EOF?
		jne	   .exit		;Nein, fertig
		move       #105,d0		;Ja, Ende
		ror	sf -,-,-		;N-FLag setzen
		rts

.exit:		clr  	sf ~-			;Flags lschen
		rts


;Job 2 "Zeichen senden"
;.............................................................................
.job2:		dec	sf d0		;Job 2?
		jeq	   .charout	;Ja

;Job 3 "String senden"
;.............................................................................
		dec	sf d0		;Job 3?			
		rts  ne			;Nein
		move	   d1,d0	;Stringpointer
.loop:		move	sf (d0),d1	;Zeichen holen
		rts  eq			;fertig wenn 0
		jsr	   .charout
		inc	   d0		;Pointer++
		jmp	   .loop

;Sendet das Zeichen in d1 
;.............................................................................
.charout:	cmp	   cts,-		;Senden erlaubt?
		jeq	   .send		;Ja
		jsr	   SWITCHTASK		;Nein, Taskwechsel und neu versuchen
		jmp	   .charout		

.send:		and	sf @statusregister,#16,- ;Sendepuffer leer?
		jne        .send		;nein, warten
		out	   d1,senderegister	;Zeichen senden
		rts

;Interroutine IRQ-Level 2
;Empfangen eines Zeichens von der seriellen Schnittstelle
;-----------------------------------------------------------------------------
seriell_int:    or         ~-,#1024,flags2		;Flags sichern
		and        @statusregister,#7,s_error	;Fehlerberprfung
		and        @empfangsregister,#$ff,s_int_ar ;Zeichen ins Arbeitsregister laden 
		clr        @statusregister		;Statusregister lschen

	        cmp	   #19,s_int_ar			;CTRL-s?
		move eq	   -,cts			;Aufhren zu Senden!
		jeq	   .ende
		cmp	   #17,s_int_ar			;CTRL-q?
		clr  eq	   cts				;Senden wieder erlaubt!
		jeq	   .ende

		move	sf (s_write),-			;Puffer frei?
		jpl        .ende			;Nein, Interruptroutine verlassen

		bswp       -,s_error,=
		or         s_error,s_int_ar,(s_write)	;Zeichen in Puffer schreiben
		inc        s_write			;Schreibpointer++
		cmp        s_write,s_buffer_end		;Pufferende?
		move eq    s_buffer,s_write		;ja, Zeiger auf Pufferanfang

		add	   -,s_write,s_int_ar		;3 Worte look-ahead fr Puffer voll
		cmp        s_int_ar,s_buffer_end	;Pufferende?
		move eq    s_buffer,s_int_ar		;ja, Zeiger auf Pufferanfang
		inc        s_int_ar			;Schreibpointer++
		cmp        s_int_ar,s_buffer_end	;Pufferende?
		move eq    s_buffer,s_int_ar		;ja, Zeiger auf Pufferanfang
		inc        s_int_ar			;Schreibpointer++
		cmp        s_int_ar,s_buffer_end	;Pufferende?
		move eq    s_buffer,s_int_ar		;ja, Zeiger auf Pufferanfang
		move	sf (s_int_ar),-			;Frei?
		jmi	   .ende

.send:		and	sf @statusregister,#16,- 	;Sendepuffer leer?
		jne        .send			;nein, warten
		out	   #19,senderegister		;CTRL-s senden
		move	   -,rts			;Flag setzen

.ende:		move   sf  flags2,~dummy		;Interrupt-Flag lschen
		jmp        seriell_int			;PC neu einstellen
