origin $8000
include lib

;=============================================================================
;SYSTEMVARIABLEN
;=============================================================================
equ bfs,6               ;Pointer auf Heap
equ efs,7               ;Adresse des 1. Wortes, welches nicht mehr zum Heap
			;gehrt (=Adresse des 1. Cachebuffers)
equ dummy,18            ;Dummyregister zur Einstellung des Statusregisters

;Serielle Schnittstelle
equ flags2,19           ;Reg. zur Sicherung der Flags
equ s_write,20          ;Schreibzeiger im Empfangspufferbereich
equ s_read,21           ;Lesezeiger im Empfangspufferbereich
equ s_int_ar,22         ;Arbeitregister der Interruptroutine
equ s_error,23          ;Fehlerregister fr Serielle-Schnittstelle,Statusregister wird im hherweritigen Byte
			;eines empfangenen Zeichens (1 Wort im Puffer) abgelegt
equ s_buffer,24         ;Zeiger auf den Pufferanfang, kann bei speziellen Kommunikationsprogrammen verndert werden
equ s_buffer_end,25     ;Zeiger auf Pufferende, erstes Wort welches nicht mehr zum Puffer gehrt 
equ cts,78              ;Variablen fr Software-Handshaking
equ rts,79

;Heap- ,Driver- und Taskverwaltung
equ h_in_use,26         ;Semaphor
equ driverlist,27       ;Pointer auf Driverliste
equ resetvector,77      ;ber diese Variable wird beim entfernen der letzten
			;Task gesprungen
;equ akt_task,17         ;Pointer auf TDB (Task-Desktriptor-Block)

;Timer: 
;equ timerh,28           ;Timer-High-Wert
;equ timerl,29           ;Timer-Low-Wert
equ flags5,30           ;Enthlt die Flags beim Interrupt
equ task_switch_tmp,31  ;Temporres Arbeitsregister beim Taskswitchen

;Konsole (Tastatur und Bildschirm)
equ umschaltflags,32    ;enthlt Flags der Umschalttasten
equ flags1,33           ;Enthlt gesicherte Flags
equ t_int_ar,34         ;Arbeitsregister der Interruptroutine
equ con_in_use,35       ;Semaphor
equ t_read_ptr,36       ;Lesezeiger
equ t_write_ptr,37      ;Schreibzeiger
equ t_buffer,38         ;Zeiger auf den Pufferanfang
equ t_buffer_end,39     ;Zeiger auf Pufferende 
equ cursor_adress,40
equ cursor_x,41
equ cursor_y,42
equ vt52_seq,43
equ save_x,44
equ save_y,45
equ save_adr,46
equ fcolor,47
equ bcolor,48

equ t_read_ptr2,49       ;Dasselbe fr die gerade nicht aktive Konsole
equ t_write_ptr2,50
equ t_buffer2,51 
equ t_buffer_end2,52
equ cursor_adress2,53
equ cursor_x2,54
equ cursor_y2,55
equ vt52_seq2,56
equ save_x2,57
equ save_y2,58
equ save_adr2,59
equ fcolor2,60
equ bcolor2,61

equ active_con,62       ;Welche Konsole (0/1?) ist aktive?
equ t_int_ar2,63        ;2. Arbeitsregister der Tastaturinterruptroutine

;SCSI:
equ status,64           ;Status
equ accesses,65         ;Zhlt Zugriffe auf Sektoren
equ media_writeprot,66  ;...
equ media_descr,67      ;Pointer auf Media-Deskriptor-Blcke (je 8 Worte)
equ not_ready_active,68 ;Zeigt an, ob die Drive-Not-ready Routine schon arbeitet
equ scsi_in_use,69      ;SCSI-Semaphor
equ scsi_buffer,70      ;Zeiger auf Arbeitspufferanfang
equ command_buffer,71   ;Zeiger auf Kommandopuffer
equ want_to_use_scsi,72 ;Fairness-Flag
equ removable,76        ;Welche Devices haben wechselbare Medien?

;Fileverwaltung
equ fdblist_in_use,73   ;Semaphor
equ fdblist,74          ;Pointer auf die FDB-Liste
equ allocate_in_use,75  ;Semaphor

;=============================================================================
;HARDWARERESET-EINSPRUNG
;=============================================================================
bs_start:       jmp        reset        ;Weiter zur eigentlichen Reset-Routine

;=============================================================================
;SYSTEMAUFRUF-DISPATCHER (Einsprungadresse $8002)
;=============================================================================
		pop        d0                   ;Funktionsnummer vom Stack
		cmp        d0,#32               ;32 ist grte Funktionsnummer
		rts mi                          ;Fertig, falls ungltige Funktionsnummer
		add        #.table,d0           ;Funktionsadresse berechnen
		add        (d0),-,=             ;Funktion anspringen

;Tabelle mit Adressen aller BS-Funktionen
.table:         dw         malloc               ;0:Speicherblock allokieren
		dw         mshrink              ;1:Speicherblock verkleinern
		dw         mfree                ;2:Speicherblock freigeben
		dw         meminfo              ;3:Freien Speicherplatz ermitteln
		dw         starttask            ;4:Task anmeldem
		dw         switchtask           ;5:Task umschalten
		dw         killtask             ;6:Task beenden
		dw         newdriver            ;7:Neuen Devicedriver anmelden
		dw         searchdriver         ;8:Devicedriver suchen
		dw         set_env              ;9:Env.-Variable setzen
		dw         search_env           ;10:Suche Environmentvariable
		dw         clr_env              ;11:Env.-Variable lschen
		dw         open                 ;12:Datei ffnen
		dw         read                 ;13:Lesen aus Datei
		dw         write                ;14:Schreiben in Datei
		dw         close                ;15:Datei schlieen
		dw         get_flags            ;16:Dateiflags ermitteln
		dw         set_flags            ;17:Dateiflags setzen
		dw         get_pos              ;18:Fileposition setzen
		dw         set_pos              ;19:Fileposition setzen
		dw         rename               ;20:Rename
		dw         delete               ;21:Delete
		dw         md                   ;22:Make Directory
		dw         rd                   ;23:Remove Directory
		dw         cd                   ;24:Change Directory
		dw         prg_load             ;25:Programm laden
		dw         prepare_search       ;26:Suchstring vorbereiten
		dw         search_next          ;27:nchsten DirEintrag suchen
		dw         scan                 ;28:Dateinamen einlesen
		dw         fehler               ;29:Fehlerstring an StandardOut ausgeben
		dw         shell                ;30:Shell
		dw         delay                ;31:Zeit verbraten
		dw         get_lng              ;32:Filelnge ermitteln

;=============================================================================
;INITIALISIERUNG
;=============================================================================
reset:          out        #0,$7ff3             ;Timer off
		move       #user_pc,0           ;Initialisierung luft im User-Mode
		move       #$fe00,~dummy        ;Switch to User-Mode

;Heapverwaltung initialisieren
user_pc:        move       #reset,resetvector   ;Resetvector einstellen
		move       #80,bfs              ;Anfang des freien Speichers
		move       #bs_start-80,(bfs)   ;Lnge dieses Speicherblocks
		add        -,bfs,d0
		clr        (d0)                 ;Speicherbereich als frei kennzeichnen
		move       #bs_start,efs        ;Ende des freien Speichers (hier=$8000)
		clr        h_in_use             ;Heapverwaltung verfgbar

;Dateiverwaltung initialisieren
		clr        fdblist              ;Semaphore fr Fileverwaltung
		clr        fdblist_in_use
		clr        allocate_in_use
		
;Taskverwaltung initialisieren, 1. Task erzeugen
		ror        -,-,sp               ;Temporrer Stack an Adresse $8000
		push       #278                 ;Wir wollen 256 Worte Environment
		jsr        malloc               ;TDB allokieren
		inc        sp
		move       d0,akt_task          ;Adresse des TDB wird Tasknummer
		dec        d0
		move       akt_task,(d0)        ;Speicherblock gehrt uns
		add        #12,d0               ;Nchster Task sind wir selbst
		move       akt_task,(d0)
		move       #.taskname,d1        ;Taskname eintragen
.nameloop:      inc        d0                   ;Name des Tasks: "XMOS 1.0"
		move    sf (d1),(d0)
		inc        d1
		jpl        .nameloop

		push       #96                  ;Speicher fr Stack
		jsr        malloc
		inc        sp
		sub        #2,d0
		add        (d0),d0,sp           ;Stackpointer init
		push       akt_task             ;Tasknummer fr Killtask
		push       #killtask            ;Killtaskadresse fr Beendigung

;Hardwaretreiber initialisieren
		push       #32          ;Driverlist init und lschen
		jsr        MALLOC
		inc        sp
		move       d0,driverlist
		sub        -,d0,d1
		sub        -,-,(d1)     ;Sys-Speicherblock (Tasknummer $ffff)
		move       #32,d1       ;Treiberliste mit Null initialisieren
.clear:         clr        (d0)
		inc        d0
		dec     sf d1
		jne        .clear

		jsr        timer        ;Timer init
		jsr        konsole      ;Diverse Treiber init
		jsr        scsi
		jsr        romdisk
		jsr        parallel
		jsr        seriell

;Standard CWD, Pfad setzen, StandardIn/Out ffnen
		push       #.defaultcwd         ;Std-Umgebungsvariablen setzen
		jsr        set_env
		push       #.defaultpath
		jsr        set_env
		push       #2                   ;Zum lesen/schreiben
		push       #.defaultio          ;Stdio ffnen
		jsr        OPEN
		add        #4,sp
		add        #20,akt_task,d1      ;In TDB eintragen
		move       d0,(d1)
		inc        d1
		move       d0,(d1)

		pushz                           ;Einschaltmeldung
		push       #.remark
		pushz
		jsr        WRITE
		add        #3,sp

		jmp        shell                ;Shell starten

.taskname:      dw      "XMOS 1.0",0,0,0,$ffff
.defaultio:     dw      "/con0",0
.defaultcwd:    dw      "cwd=/b/",0
.defaultpath:   dw      "path=/rom/",$3b,"/b/",0
.remark:        dw 10,13,"Welcome to the "
		dw $1b,$47,$00,$00,$1f,$3e,$7c,$f8,$7c,$3e,$1f,$3e,$7c,$f8,$7c,$3e,$1f,$00
		dw $1b,$47,$00,$00,$7c,$f8,$70,$20,$00,$00,$00,$00,$00,$20,$70,$f8,$7c,$00
		dw "Xilinx Multitasking Operating System V1.0, November 1993",13,10,0

;Komponeneten einbinden
include konsole         ;Konsolentreiber, Tastaturinterruptroutine, Zeichensatz
include scsi            ;SCSI-Treiber und Cache
include shelltxt        ;Shell
include seriell         ;Text-Treiber fr serielle Schnittstelle
include parallel        ;Parallelport-Treiber
include file_sub
include file            ;Dateiverwaltung
include romdisk         ;ROM-Disk Treiber mit Datenbereich


;=============================================================================
;GERTETREIBER-VERWALTUNG
;=============================================================================
;-----------------------------------------------------------------------------
;NEWDRIVER
;-----------------------------------------------------------------------------
;Anmelden eines neuen Treibers
;d0: Zeiger in der Treiberliste
;d1: Zhler
newdriver:      irqoff                          ;Keine Unterbrechung zulassen
		move       driverlist,d0        ;Suche freien Eintrag
		move       #32,d1               ;Zhler, max. 32 Eintrge
.listsearch:    dec     sf d1
		jmi        .full                ;Driverliste leider voll
		cmp        (d0),-               ;Eintrag frei?
		inc  ne    d0                   ;Nein weiter
		jne        .listsearch

;d1: Zeiger im Stack 
		add        -,sp,d1              
		move       (d1),(d0)    ;Treiberanfangsadresse eintragen
.ende:          irqon
		rts

.full:          ror     sf -,-,-        ;N-Flag setzen, Liste ist voll
		jmp         .ende

;-----------------------------------------------------------------------------
;SEARCHDRIVER
;-----------------------------------------------------------------------------
;Suchen nach einem angemeldeten Treiber
;d0, d1: Zeiger auf zu suchenden Treibernamen
;d2: Zeiger auf Treiberliste
;d3: Zeiger auf Treibernamen im Treiber
searchdriver:   add        -,sp,d0
		push       d2
		push       d3
		move       (d0),d0              ;Pointer auf Name des gesuchten Drivers
		move       driverlist,d2        ;Adresse der Treiberliste
.loop:          move       d0,d1
		move    sf (d2),d3              ;Adresse des Treibers holen
		jeq        .not_found           ;Es gibt keine Treiber mehr
		cmp        (d1),(d3)            ;1.Zeichen gleich?
		jne        .notyet              ;Nein, Treiber ist nicht der gesuchte
		inc         d1
		inc         d3
		cmp        (d1),(d3)            ;2. Zeichen gleich?
		jne        .notyet              ;Nein, Treiber ist nicht der gesuchte
		cmp        (d1),-               ;Ende des Namens?
		jeq        .found
		inc         d1
		inc         d3
		cmp        (d1),(d3)            ;3. Zeichen gleich?
		jne        .notyet              ;Nein, Treiber ist nicht der gesuchte
		cmp        (d1),-               ;Ende des Namens?
		jeq        .found
		inc         d1
		inc         d3
		cmp        (d1),(d3)            ;4. Zeichen  gleich?
		jne        .notyet              ;Nein, Treiber ist nicht der gesuchte
.found:         add        #5,(d2),d0           ;Treiber gefunden, Adresse in d0
		clr     sf ~-                   ;N-Flag lschen
.end:           pop        d3
		pop        d2
		rts

;Dieser Treiber war unpassend, nchsten Treiber betrachten.
.notyet:        inc        d2                   ;Nchster Treiber
		jmp        .loop

;Liste zu Ende, Treiber nicht gefunden.
.not_found:     move       #106,d0              ;Device exitiert nicht!
		ror     sf -,-,-                ;N-Flag setzen
		jmp        .end

;=============================================================================
;TASKVERWALTUNG
;=============================================================================
;-----------------------------------------------------------------------------
;STARTTASK
;-----------------------------------------------------------------------------
;Task im System anmelden und TDB erzeugen
;d0, d2: Adresse der Task
starttask:      add        #2,sp,d0
		move       (d0),d0      ;Startadresse der Task holen
		jsr        find_block   ;zugehrigen Speicherblock ermitteln
		rts  mi
		push       d2
		move       d0,d2        ;Speicherblockadresse retten            

		add        #2,sp,d0
		add        #22,(d0),d0  ;Gre des Environmentbereiches

;Speicher fr TDB mit Environmentbereich anfordern
		push       d0           ;Neuen TDB anlegen
		jsr        MALLOC
		inc        sp
		pop  mi    d2
		rts  mi                 ;Bei Fehler: Task kann nicht gestartet werden

;d0: Adresse des TDB
;d1: Adresse des zweiten Verwaltungswortes dieses Speicherblocks (Besitzer).            
		sub        -,d0,d1      ;Speicherblock gehrt dem neuen Task
		move       d0,(d1)
		add        #3,sp,d1     ;d1: Startadresse der neuen Task holen...
		move       (d1),(d0)    ;...und in TDB eintragen

;Taskname eintragen
;d0: Zeiger in TDB
;d1: Zeiger auf Tasknamen
;d2: Zhler
		push       d2
		inc        d1           ;Taskname in TDB eintragen
		move       (d1),d1      ;Pointer auf Taskname holen
		add        #12,d0
		move       #8,d2        ;8 Zeichen kopieren
.copyloop:      move    sf (d1),(d0)
		inc  ne    d1
		inc        d0
		dec     sf d2
		jne        .copyloop
		pop        d2

;StandardIn und StandardOut-FDB eintragen und diese Task als Besitzer dort eintragen
;d0: Zeiger in TDB
;d1: Zeiger auf StdIn im Stack
;d2: Adresse der Task
		add        #5,sp,d1     ;Stdin verwalten
		move       (d1),d1
		move    sf d1,(d0)
		dec        d1
		sub ne     #20,d0,(d1)  ;Stdin-FDB gehrt jetzt uns
		inc        d0
		add        #6,sp,d1     ;Ebense Stdout
		move       (d1),d1
		move    sf d1,(d0)
		dec        d1
		sub ne     #21,d0,(d1)

;Environmentbereich der Erzeugertask kopieren.
;Diese Task erbt den Env.-Bereich
;d1: Zeiger auf Ende des TDB
;d2: Zeiger auf Environmentbereich der aufrufenden Task
		push       d0
		push       d2
		inc        d0                   ;d0 zeigt auf neues Environment 
		sub        #24,d0,d1
		add        (d1),d1              ;d1 zeigt auf Ende des TDB
		add        #22,akt_task,d2      ;d2 zeigt auf aktuelles Environment
		cmp        d0,d1                ;Environment leer?
		jeq        .noenv               ;Ja, nichts kopieren
.copyenv:       move       (d2),(d0)            ;Environment vererben
		inc        d2
		inc        d0
		cmp        d0,d1
		jne        .copyenv
		dec        d0                   ;Environmentende setzen
		clr        (d0)
.noenv:         pop        d2
		pop        d0           

;gerettet Falgs mit $ffff vorbesetzen
		sub        #20,d0
		sub        -,-,(d0)             ;Flags sind $ffff

;Stackpointer auf Ende der Task einstellen.
;Auf den Stack Tasknummer und Adresse der Killtaskroutine pushen
;d1: Zeiger auf Ende der Task (Stack)
;d2: Zeiger auf Anfangsadresse der Task
		add        (d2),d2,d1           ;d1 ist Stackpointer der neuen Task
		dec        d1
		sub        -,d0,(d1)            ;Eigene Tasknummer auf den Stack pushen
		inc        d2
		move       (d1),(d2)            ;Programmspeicherblock gehrt der neuen Task
		dec        d1                   ;ebenso Taskrcksprungadresse 
		move       #KILLTASK,(d1)
		inc        d0                   
		move       d1,(d0)              ;Stackpointer in TDB eintragen

;TDB in Taskliste einhngen, darf nicht unterbrochen werden
;d1: Zeiger auf TDB des Nachfolgers
		irqoff
		add        #9,d0                ;Nachfolger der aktuellen...
		add        #11,akt_task,d1      ;...Task wird Nachfolger... 
		move       (d1),(d0)            ;...der neuen Task
		sub        #11,d0,(d1)          ;Neue Task wird Nachfolger der aktuellen Task
		irqon

		clr     sf ~-                   ;N-Flag lschen
		pop        d2
		rts

;-----------------------------------------------------------------------------
;SWITCHTASK
;-----------------------------------------------------------------------------
;Taskwechselroutine, kann nicht unterbrochen werden.
;Die Routine ist um die Geschwindigkeit zu optimieren linear und ohne
;Schleifen programmiert. Die Routine luft im Interruptlevel 5.
switchtask:     irqoff                                  ;Unterbrechung verhindern
		or         ~-,#256,flags5               ;Flags retten, IRQ enable
		move       #.do_it,5
		and        ~-,#$dfff,~dummy             ;In IRQ5 schalten
.do_it:         move    sf task_switch_tmp,(akt_task)   ;Autoswitch: User-PC retten
		move eq    (sp),(akt_task)              ;Syscall: User-PC
		add        -,akt_task,task_switch_tmp
		move       flags5,(task_switch_tmp)     ;Flags retten
		inc        task_switch_tmp
		move       sp,(task_switch_tmp)         ;Stackpointer retten
		inc        task_switch_tmp
		move       d0,(task_switch_tmp)         ;Arbeitsregister retten
		inc        task_switch_tmp
		move       d1,(task_switch_tmp)
		inc        task_switch_tmp
		move       d2,(task_switch_tmp)
		inc        task_switch_tmp
		move       d3,(task_switch_tmp)
		inc        task_switch_tmp
		move       d4,(task_switch_tmp)
		inc        task_switch_tmp
		move       d5,(task_switch_tmp)
		inc        task_switch_tmp
		move       d6,(task_switch_tmp)
		inc        task_switch_tmp
		move       d7,(task_switch_tmp)
		inc        task_switch_tmp

;Nchste Task aktivieren
		move       (task_switch_tmp),task_switch_tmp    ;Nchste Tasknummer holen
		move       task_switch_tmp,akt_task
		move       (task_switch_tmp),0          ;USER-PC der neuen Task
		inc        task_switch_tmp
		move       (task_switch_tmp),flags5     ;Flags zurck
		inc        task_switch_tmp
		move       (task_switch_tmp),sp         ;Stackpointer zurck
		inc        task_switch_tmp
		move       (task_switch_tmp),d0         ;Arbeitsregister zurck
		inc        task_switch_tmp
		move       (task_switch_tmp),d1
		inc        task_switch_tmp
		move       (task_switch_tmp),d2
		inc        task_switch_tmp
		move       (task_switch_tmp),d3
		inc        task_switch_tmp
		move       (task_switch_tmp),d4
		inc        task_switch_tmp
		move       (task_switch_tmp),d5
		inc        task_switch_tmp
		move       (task_switch_tmp),d6
		inc        task_switch_tmp
		move       (task_switch_tmp),d7
		clr        task_switch_tmp
		move    sf flags5,~dummy        ;Flags zurck

;-----------------------------------------------------------------------------
;TIMER-IRQ-ROUTINE
;-----------------------------------------------------------------------------
;Die Timerroutine incrementiert die Zeitzhler lt einen Taskwechsel aus.
timer_irq:      out        -,$7ff3              ;Timer-IRQ-Besttigung
		or         ~-,#8192,flags5      ;Flags retten
		inc     sf timerl               ;Timer-Low inkrement
		inc  cs    timerh               ;Timer-High inkrement
		cmp        task_switch_tmp,-    ;Bei Strungen auf der IRQ-Leitung: Kein Autotaskswitch
		move eq    0,task_switch_tmp    ;Autotaskswitch prepare
		move eq    #switchtask,0        ;Autotaskswitch
		move    sf flags5,~dummy        ;Flags zurck
		jmp        timer_irq            ;u.s.w.


;-----------------------------------------------------------------------------
;KILLTASK
;-----------------------------------------------------------------------------
;Killtask entfernt einen Task aus dem System
;Speicher wird zurckgegeben und Dateien geschlossen
killtask:       irqoff                          ;Keine Unterbrechung
		add        -,sp,d0              
		move       (d0),d0              ;d0 zeigt auf TDB des zu lschenden Tasks

		cmp        akt_task,d0          ;Ist dies ein Selbstmord?
		jeq        .suicide             ;ja

;Ist eine gltige Tasknummer angegeben?
;d0: gesuchte Tasknummer
;d1: Zeiger in TDB-Liste
		move       akt_task,d1          ;Suche nach Vorgngertask
.gueltig:       add        #11,d1
		cmp        d0,(d1)              ;Ist nchste Task unsere gesuchte Task?
		jeq        .found               ;ja
		move       (d1),d1
		cmp        d1,akt_task          ;Haben wir schon alle Tasks durchgeschaut?
		jne        .gueltig
		move       #101,d0              ;ja
.fehler:        irqon
		ror     sf -,-,-                ;N-Flag setzen, Tasknummer war ungltig
		rts

;Tasknummer ist gltig. User-Pc der zu entfernenden Task auf die Killtaskroutine
;einstellen und die Tasknummer auf deren Stack pushen. Beim nchsten Taskwechsel
;begeht diese Routine einen Selbstmord
.found:         move       #killtask,(d0)       ;User PC auf Killtaskroutine stellen
		add        #2,d0,d1             ;Adresse des Stackpointers
		move       (d1),d1              ;Wert des Stackpointers
		move       d0,(d1)              ;Tasknummer auf Stack ablegen
		add        #2,d0,d1             ;Adresse des Stackpointers
		dec        (d1)
		irqon
		clr     sf ~-
		rts

;Selbstmord einer Task, Semaphore lschen
.suicide:       cmp        akt_task,h_in_use    ;benutzt Task die Heapverwaltung?
		clr eq     h_in_use             ;Ja, Heapverwaltung freigeben
		cmp        akt_task,con_in_use  ;...ebenso Konsole...
		clr eq     con_in_use
		cmp        akt_task,fdblist_in_use      
		clr eq     fdblist_in_use
		cmp        akt_task,allocate_in_use
		clr eq     allocate_in_use
		cmp        akt_task,scsi_in_use
		irqon
		jne	   .fdbstart
		out        -,scsi_ctrl		;SCSI-Bus Reset
		push	   #2			;kurz warten
		jsr	   DELAY
		clr        @scsi_ctrl		;SCSI-Bus Reset clear
		clr        scsi_in_use

;geffnete Dateien ordnungsgem schlieen
;d1: Zeiger auf FDB-Liste
.fdbstart:      move       #fdblist,d1          ;Alle geffneten Dateien des Tasks lschen
.fdblist:       cmp        (d1),-
		jeq        .go_on
		move       (d1),d1
		dec        d1
		cmp        (d1),akt_task        ;gehrt die Datei dieser Task?
		inc        d1
		jne        .fdblist
		push       d1                   ;Datei schlieen
		jsr        CLOSE
		inc        sp
		jmp        .fdbstart

;RESET auslsen?
.go_on:         irqoff                          ;Unterbrechung unterbinden
		add        #11,akt_task,d0      ;Ist dies die letzte Task?
		cmp        akt_task,(d0)
		add  eq    resetvector,-,=      ;Reset auslsen

;Task aus der Taskliste entfernen
;d1: Zeiger auf Vorgngertask
		move       akt_task,d1          ;Suche nach Vorgngertask
.search:        add        #11,d1
		cmp        akt_task,(d1)        ;Ist nchste Task unsere gesuchte Task?
		move ne    (d1),d1
		jne        .search

		add        #11,akt_task,d0      ;Task aus Taskliste streichen
		move       (d0),(d1)

;Durch Task belegt Speicherbereich freigeben
;d1: Zeiger auf Speicherliste
		move       bfs,d1
.heapsearch:    inc        d1           ;Ist Speicherblock durch diese Task belegt?
		cmp        (d1),akt_task
		clr eq     (d1)         ;Ja, Speicherblock freigeben
		dec        d1
		add        (d1),d1      ;nchster Block
		cmp        d1,efs       ;letzter Block?
		jne        .heapsearch  ;Noch nicht

		jsr        mfree.melt   ;...Garbage-Collection in Speicherliste
		irqon
		jmp        switchtask   ;Ja, zur nchsten Task


;=============================================================================
;HEAPVERWALTUNG
;=============================================================================
;-----------------------------------------------------------------------------
;MALLOC
;-----------------------------------------------------------------------------
;Arbeitsspeicher zur Verfgung stellen
;d0: Zeiger auf Speicherliste
;d1: Zeiger auf angeforderte Speichergre
;d2: Gesuchte Speichergre (+2 Worte Verwaltung)
;d3: Adresse des letzten freien Blocks
malloc:         irqoff                          ;Semaphor berprffung darf nicht unterbrochen werden
		move    sf h_in_use,-           ;Heapverwaltung schon aktiv?
		move eq    akt_task,h_in_use    ;Falls nicht, belegen
		irqon
		jeq        .semok
;(Hier knnte ein Fairnessflag gesetzt werden!)
		jsr        SWITCHTASK
		jmp        malloc               ;Sonst warten

.semok:         push       d2
		push       d3
.retry:         move       bfs,d0       ;Adresse des 1. Blocks holen
		add        #3,sp,d1
		add        #2,(d1),d2   ;Soviel will der User (+2 Worte fr Verwaltung)

;Block der angeforderten Gre suchen           
.search:        cmp        d2,(d0)      ;Ist aktueller Block gro genug?
		jmi        .toosmall    ;Nein
		add        -,d0,d1      ;Ja, aber ist er frei?
		move    sf (d1),-
		jeq        .found       ;Auch das -> Block gefunden

.toosmall:      move       d0,d3        ;Adresse des letzten Blocks merken
		add        (d0),d0      ;Pointer auf nchsten Block...
		cmp        efs,d0       ;Gibt es gar keinen mehr?
		jne        .search      ;Doch, weitersuchen

;.............................................................................
;Auch der letzte Block war zu klein oder belegt -> Speicher vom SCSI-Cache
;zurckfordern
;Achtung! Hier ist eine Verbindung zum SCSI-Treiber!
		cmp        (d1),-       ;Ist dieser letzte Block frei?
		sub eq     (d3),d2      ;Ja, nur den fehlenden Speicher vom Cache anfordern
		add        efs,d2
		cmp        d2,#bs_start-6*260 ;Wrde der Speicher reichen?
		move mi    #103,d0
		jmi        .errorend    ;Nein, Speicheranforderung nicht erfllbar

;Treiberaufruf an SCSI um Cache freizugeben!
		push       d2           ;Sonst: Dec-Buffer-Job fr einen SCSI-Driver aufbauen
		push       #5           
		jsr        scsidriver0+5
		add        #2,sp
		add        -,d3,d0      ;Falls letzter Block belegt war, neuen Block anlegen
		cmp        (d0),-
		add ne     (d3),d3
		sub        d3,efs,(d3)
		inc        d3
		clr        (d3)         ;Block ist frei
		jmp        .retry       ;und nochmal versuchen
;.............................................................................
;Erst ab 16 verbleibenden Worten wird ein Block geteilt um keinen Verwaltungs-
;Overhead hervorzurufen.
;d1: Zeiger auf die Verwaltungsworte
;d3: Gre des freien Speicherblocks
.found:         sub        #16,(d0),d3          ;Soll Block geteilt werden?
		cmp        d2,d3
		jmi        .allocateblock       ;Nein, nur belegen

		move       (d0),d3              ;Bisherige Blockgre holen
		move       d2,(d0)              ;Block wird so gro wie angefordert wurde
		add        d2,d0,d1             ;Adresse des neu entstandenen Blocks
		sub        d2,d3,(d1)           ;Dessen neue Gre eintragen
		inc        d1
		clr        (d1)                 ;neuer Block ist frei

;Als Blockbesitzer ist aktuelle Task eintragen
.allocateblock: inc        d0
		move       akt_task,(d0)        ;Aktuelle Tasknummer eintragen
		inc        d0                   ;Adresse des freien Bereiches
		clr     sf -                    ;N-Flag lschen

.errorend:      pop        d3
		pop        d2
		clr        h_in_use             ;Semaphor lschen
;(Hier mte Fairnessberprfung eingesetzt werden!)
		rts

;-----------------------------------------------------------------------------
;MFREE
;-----------------------------------------------------------------------------
;MFREE gibt Speicher frei
mfree:          irqoff                          ;Kein Unterbrechung bei Semaphorbehandlung
		move    sf h_in_use,-           ;Heapverwaltung schon aktiv?
		move eq    akt_task,h_in_use    ;Falls nicht, aktivieren
		irqon
		jeq        .semok
;(Hier knnte ein Fairnessflag gesetzt werden!)
		jsr        SWITCHTASK
		jmp        mfree        ;Sonst warten

.semok:         add        -,sp,d0
		sub        #2,(d0),d0   ;Adresse des freizugebenden Blocks

;Speicherblock suchen
;d0: Adresse des freizugebenden Blocks
;d1: Zeiger auf Speicherliste
		move       bfs,d1
.suchloop:      cmp        d1,d0        ;Block im Heap gefunden?
		jeq        .free        ;Ja, freigeben
		add        (d1),d1      ;Nchster Block
		cmp        d1,efs       ;War das der letzte?
		jne        .suchloop    ;Noch nicht
.illegal:       ror     sf -,-,-        ;Blockadresse war ungltig
		move       #101,d0
		jmp        .ende

;Gehrt zu lschender Block der Task?
.free:          inc        d0
		cmp        (d0),akt_task ;Eigener Block?
		jne        .illegal     ;Nein, nicht lschen
		clr        (d0)         ;Block freigeben

;Hier werden nun 2 aufeinanderfolgende freie Blcke zu einem verschmolzen
;.............................................................................
.melt:          move       bfs,d0       ;d0: Adresse des 1. Blocks
.meltloop:      add        d0,(d0),d1   ;d1: Adresse des 2. Blocks
		cmp        d1,efs       ;Gibt es den 2. Block gar nicht?
		jeq        .inc_buffer
		inc        d0           ;Zeiger auf Block-Frei-Wort
		cmp        (d0),-       ;1. Block frei?
		dec        d0
		jne        .notfree     ;nein
		inc        d1           ;Zeiger auf Block-Frei-Wort
		cmp        (d1),-       ;2. Block frei?
		dec        d1
		add eq     (d1),(d0)    ;Ja: Lnge des 1. Blocks wird Summe der beiden Blocklngen
.notfree:       add ne     (d0),d0      ;1. Zeiger weiterschalten
		jmp        .meltloop

;.............................................................................
;Hier ist eine Verbindung der Speicherverwaltung mit dem SCSI-Treiber!
;Falls letzter Block sehr gro ist, ein Teil davon als Cachebuffers freigeben
.inc_buffer:    cmp        (d0),#3000   ;Letzter Block sehr gro?
		jpl        .ende        ;Nein, keinen neuen Cachebuffer einrichten

		sub        #257,efs     ;Neuer Buffer
		clr        (efs)        ;Zugriffzhler=0
		sub        #3,efs
		clr        (efs)        ;Buffer ist frei
		sub        #260,(d0)    ;Block ist um soviel kleiner
		jmp        .inc_buffer
;.............................................................................
.ende:          clr        h_in_use
;(Hier mte Fairnessberprfung eingesetzt werden!)
		rts

;-----------------------------------------------------------------------------
;MSHRINK
;-----------------------------------------------------------------------------
mshrink:        irqoff                          ;Keine Unterbrechung
		move    sf h_in_use,-           ;Heapverwaltung schon aktiv?
		move eq    akt_task,h_in_use    ;Falls nicht, aktivieren
		irqon
		jeq        .semok
;(Hier knnte ein Fairnessflag gesetzt werden!)
		jsr        SWITCHTASK
		jmp        mshrink      ;Sonst warten

.semok:         add        -,sp,d1
		sub        #2,(d1),d0   ;Adresse des Blocks holen

;Block suchen
;d0: Adresse des zu verkleinernden Blocks
;d1: Zeiger auf Speicherliste
		move       bfs,d1
.suchloop:      cmp        d1,d0        ;Block im Heap gefunden?
		jeq        .shrink      ;Ja, verkleinern
		add        (d1),d1      ;Nchster Block
		cmp        d1,efs       ;War das der letzte?
		jne        .suchloop    ;Noch nicht
.illegal:       ror     sf -,-,-        ;Blockadresse war ungltig
		move       #101,d0
		jmp        mfree.ende

;Block zusammenschmelzen
;d1: neue Blockgre
.shrink:        add        #2,sp,d1
		move       (d1),d1      ;neue Blockgre holen
		inc        d0
		cmp        (d0),akt_task ;Eigener Block?
		jne        .illegal     ;Nein, nicht verkleinern
		dec        d0

		add        #3,d1        ;Soll Block geteilt werden?
		cmp        (d0),d1
		sub        #3,d1
		jpl        mfree.ende   ;Nein, der entstehende Block wre zu klein

;d2: Gre des neu entstandenen Blocks
		push       d2
		sub        d1,(d0),d2
		add        #2,d1,(d0)   ;Neue Gre des verkleinerten Blocks
		add        (d0),d0
		sub        #2,d2,(d0)   ;Gre des neu entstehenden Blocks
		pop        d2
		inc        d0
		clr        (d0)         ;Neuer Block ist frei
		jmp        mfree.melt   ;Meltloop aufrufen

;-----------------------------------------------------------------------------
;FIND_BLOCK
;-----------------------------------------------------------------------------
;Nur interner Aufruf.
;Findet zu einer Adresse in d0 den zugehrigen Speicherblock
;d0: Zeiger auf Speicherliste
;d1: Adresse des zu suchenden Blocks
;d2: Speicherblockadresse gegen die geprft wird
find_block:     push       d1
		push       d2
		move       d0,d1        ;Adresse im zu suchenden Block
		move       bfs,d0       ;Heapliste
.loop:          add        (d0),d0,d2   ;Ist aktueller Block der gesuchte?
		cmp        d1,d2
		jpe        .found
		add        (d0),d0      ;Nchsten Puffer betrachten
		cmp        efs,d0       ;letzter Block?
		jmi        .loop        ;Nein, weitersuchen
		ror     sf -,-,-        ;Speicherblock nicht gefunden
.found:         pop        d2
		pop        d1
		rts

;-----------------------------------------------------------------------------
;MEMINFO
;-----------------------------------------------------------------------------
Ergebnisse:
;d0: Lnge des grten Speicherblocks
;d1: Lnge des freien Speichers insgesamt
;Registerverwendung:
;d2: Zeiger auf Speicherliste
meminfo:        push       d2           ;Ermittelt noch zur Verffgung stehenden Speicher
		push       d3
		push       d4
		push       d5
		push       d6
		clr        d0
		clr        d1
		clr        d5
		move       bfs,d2       ;Zeiger auf ersten Speicherblock

;Freie Blcke suchen
;d3: Adresse des 2. Verwaltungswortes (Besitzer)
;d4: Adresse eines freien Blocks
;d5: Adresse des letzten Blocks der Liset
;d6: Adresse des nchsten Blocks
.loop:          add        -,d2,d3
		cmp        (d3),-       ;Speicherblock frei?
		jne        .next

;d3: Lnge des Blocks
		sub        #2,(d2),d3
		add        d3,d1        ;ja,Gesmatspeicherfreizhler erhhen
		cmp        d3,d0        ;Gefundener Block grer als bisher grter
		move mi    d3,d0        ;ja, eintragen
		move mi    d2,d4        ;Adresse des Blocks merken              add        d2,(d2),d6
		cmp        d6,efs       ;Ist dieser Block der letzte freie?
		move eq    d3,d5        ;ja, Lnge merken

.next:          add        (d2),d2
		cmp        d2,efs       ;Ende des Speichers erreicht?
		jpe        .loop

;.............................................................................
;Speicherplatz bis zum Minimal-Cache hinzu addieren
;Achtung! Verbindung der Speicherverwaltung mit dem SCSI-Treiber (Cache)!
;d2: Verbleibender Speicher ohne Cache
;d3: Adresse des letzten freien Blocks
		sub        efs,#bs_start-6*260,d2       ;Speicherplatz ohne Cache-Puffer
		add        d2,d1                        ;Insgesamt freier Speicher
		add        d4,(d4),d3
		cmp        d3,efs       ;Schlo der lngste freie Block an efs an?
		add eq     d2,d0        ;ja, dann Cache-Speicher hinzuaddieren
		jeq        .ende

		add        d5,d2
		cmp        d2,d0        ;Ist der letzte freie+Cache-Speicher grer als der bisher grte Block?
		move mi    d2,d0        ;Ja
;.............................................................................

.ende:          pop        d6
		pop        d5
		pop        d4
		pop        d3
		pop        d2
		clr     sf ~-
		rts

;=============================================================================
;TIMER                                     
;=============================================================================
;Der Timer-Interrupt wird 8000000/(2^19)-mal in der Sekunde ausgelt.

;-----------------------------------------------------------------------------
;INITIALISIERUNG
;-----------------------------------------------------------------------------
timer:          clr        timerl       ;Register lschen
		clr        timerh
		clr        task_switch_tmp
		move       #timer_irq,5 ;Interruptroutine einstellen
		out        -,$7ff3      ;Timer on
		rts                              

;-----------------------------------------------------------------------------
;DELAY
;-----------------------------------------------------------------------------
;d0: Zeitpunkt der Rckkehr, low
;d1: Zeitpunkt der Rckkehr, high
delay:          add        -,sp,d0              ;Timer-hi und -lo holen und...
		irqoff                          ;... gewnschte Wartezeit
		add     sf timerl,(d0),d0       ;addieren
		addc       timerh,-,d1
		irqon

.wait_hi:       cmp        d1,timerh            ;Warten, bis timer-high... 
		jpl        .wait_low            ;geforderten Wert erreicht hat
		jsr        switchtask
		jmp        .wait_hi

.wait_low:      cmp        d0,timerl            ;ebenso fr timer-low
		rts pl
		jsr        SWITCHTASK
		jmp        .wait_low


dw "Hardware und Software entwickelt von Uwe Reinhardt und Oliver Henning, 1993"
 
